/* 
 *  FFTGRAF Benchmark Version 3.0 Copyright (c) Roy Longbottom 2004
 *
 *  Version 1 Mar 2001
 *  Version 1.01 Apr 2001, added compare, no default graph
 *  Version 1.02 May 2001, added compare text output
 *  Version 1.1  Jan 2002, different display method
 *  Version 2.0  Jan 2002, Burst reads and cache segments + assembler
 *  Version 3.0  Oct 2002, SSE, SSE2 and 3DNow
 * 
 *  FFTGraf benchmark uses optimised FFT code written by Scott Taylor.
 *  It produces a graph of results besides text output in file FFTGraf.txt.
 *  The text file can be read to reproduce graphs on the screen.
 * 
 *  The data for FFTs is not loaded or stored from sequential memory
 *  addresses for much of the time. As the hardware loads this data in 32
 *  or 64 byte bursts, much of it is redundant, resulting in slow
 *  performance. AMD and Intel P4 CPUs suffer most as their bursts are
 *  64 bytes.
 *
 *  Version 2 of the program produces significant speed improvements 
 *  mainly by making more efficient use of caches and using all burst 
 *  data. Scotts new code divides the data from RAM into L2 cache sized 
 *  segments, leading to a 2 x improvement on large FFTs. Roys new code 
 *  unrolls the critical loop to make effective use of burst data read 
 *  into caches, leading to up to a further 2 x improvement or more with 
 *  PCs that use 64 byte bursts. Roy has also supplied assembly code for 
 *  the main calculations which helps a little.  
 *
 *  Version 3 includes additional code from Roy to use SSE or 3DNow
 *  instructions for single precision calculations and SSE2 instructions
 *  for double precision calculations, when these are provided by the
 *  CPU. Pentium III and AMD XP CPUs have SSE and Pentium 4 has SSE
 *  and SSE2. AMD CPUs have 3DNow. The latest AMD Athlon 64 also has
 *  SSE and SSE2.
 *
 *  The program calculates 1024 to 1048576 sized FFTs using both single 
 *  and double precision numbers. Each can be calculated up to 12 times 
 *  (default 5). Running time is calculated in milliseconds but graphs 
 *  are plotted using milliseconds per K size for each run. Although ms/K 
 *  increases as FFTs become larger, a distinct difference is usual when 
 *  data is too large for a cache. The ms/K scale is self calibrating and 
 *  increases as the tests run. 
 *
 *  Version 1 memory demands in bytes were up to 16 times the FFT size 
 *  on single precision and 28 times on double precision. These also 
 *  apply to Versiona 2 and 3 up to size 8192. Above this, they can be 
 *  up to 28 times on SP and 52 times on DP - maximum 52 MB. Data spanned 
 *  in the critical timing loop is 8 or 16 times FFT size at 8192 and 
 *  below and 16 x SP or 32 x DP above 8192. 
 *
 *  Version 3 needs Microsoft Visual C++ 6.0 Pro with Service Pack 5
 *  and Processor Pack, which is required to use the new instructions.
 *  These are used via assembly code.
 *
 *  Modules
 *
 *  Config.cpp - via WhatConfig() identifies CPU type/speed, supported
 *             instruction extensions, memory size and OS version,
 *             also getStartTime(), getRunSecs(), local_time()
 *
 *  FFTGraf3.rc - menus and dialog boxes
 *        MENU_ABOUT      - AboutDlgProc()
 *        MENU_HELP      -  AboutDlgProc()
 *        MENU_PARAM      - WriteReadGraph()
 *        MENU_SCALE      - WriteReadGraph()
 *        MENU_RUNGRAPH   - Run tests doRun = 1
 *        MENU_LOAD       - Load data from log file - GetFileName(), loadFile()
 *        MENU_COMPARE    - Load two logs as MENU_LOAD, Compare0(), Compare1(), Compare2()
 *        MENU_SAVE       - GetFileName(), saveBmp()
 *        MENU_COMPAREOUT - compareTxt()
 *        MENU_VIEWGRAPH  - reView()
 *        MENU_EXIT       - CleanUpAndPostQuit()
 *
 *  FFTGraf.h and FFTGraf.ico
 *
 *  FFTGraf3.cpp
 *
 *  Particularly with Windows XP, on running a benchmark which displays
 *  on-going results, the window for the latter can be covered by another
 *  window. To overcome this, all tests are run via the WinMain message
 *  processing loop using doRun and jpass variables. 
 *
 *  The main window is used to display all information and graphs.
 *  WindowProc() creates 3 buttons. RUN_BUTTON sets doRun = 1, PARAM_GRAPH
 *  executes WriteReadGraph() for variables and FINISH_GRAPH is exit via
 *  CleanUpAndPostQuit(). The graphs are drawn by drawGraphs() via reView()
 *  which is used after each test or when anything has changed.
 * 
 *  Functions
 *
 *  WinMain() - usual Windows functions + allocates memory and runs tests
 *          doRun = 1 - SPCalculate(), checkNoiseSP() for each jpass
 *          doRun = 2 - DPCalculate(), checkNoiseDP() for each jpass
 *          saveGrafTxt(), reView(), free memory on exit
 *  InitApp() - usual Windows functions, allocates variable values
 *  WindowProc() - usual Windows functions for msgs, buttons & menus
 *  AboutDlgProc() - displays Help and About
 *  CleanUpAndPostQuit() - set up exit
 *  WriteReadGraph() - dialog box for setting run time parameters, UpdateCombos()
 *  UpdateCombos() - update dialog box settings
 *  reView() - drawGraphs()
 *  drawGraphs() - display text and graphs
 *  saveGrafTxt() - save results in log file FFTGraf.txt
 *  saveBmp() - save a copy of graph in BMP format
 *  loadFile() - load selected data from a log file 
 *  GetFileName() - standard procedure
 *  Compare0() - zeroise variables
 *  Compare1() - data for first to compare
 *  Compare2() - data for second to compare, drawGraphs()
 *  compareTxt() - save comparison bar chart in Compare2.txt  
 *
 *  SPCalculate() - bitrev(), createExpTableSP(), createComplexDataSP(),
 *        getStartTime(), fftSP() or fftSP2 depending on cache size parameter
 *        TsizeLL, getRunSecs(), drawGraphs() 
 *  bitrev(int) - bit reverse table
 *  createExpTableSP() - create exponent table
 *  createComplexDataSP () -  create data to avoid errors
 *  fftSP() -  Break the work up into NumSeg work units of size TsizeL2,
 *        ass1SP3DNow() or ass1SP(), ass2SPSSE() or ass2SP3DNow() or ass2SP(),
 *        ass4SPSSE() or ass4SP3DNow() or ass4SP()
 *        if necessary ass4SPSSE() or ass4SP3DNow() or ass4SP()   
 *  fftSP2() - ass1SP3DNow() or ass1SP(), ass2SPSSE() or ass2SP3DNow() or
 *        ass2SP(), ass4SPSSE() or ass4SP3DNow() or ass4SP()
 *  DPCalculate() - bitrev(), createExpTableDP(), createComplexDataDP(),
 *        getStartTime(), fftDP() or fftDP2 depending on cache size parameter
 *        TsizeLL, getRunSecs(), drawGraphs() 
 *  createExpTableDP() - create exponent table
 *  createComplexDataDP() - create data to avoid errors
 *  fftDP() - Break the work up into NumSeg work units of size TsizeL2,
 *        ass1DPSSE2() or ass1DP(), ass2DPSSE2() or ass2DP(), ass4DPSSE2()
 *        or ass4DP(), if necessary ass4DPSSE2() or ass4DP()
 *  fftDP2() - ass1DPSSE2() or ass1DP(), ass2DPSSE2() or ass2DP(),
 *             ass4DPSSE2() or ass4DP()
 *
 *  Note: the ass1, ass2, ass4 functions could be combined. These are derived
 *  from Version 1 C code (see Version 1 was at end and comments in assembly
 *  code) as is and unrolled for loading more data on burst reads from memory.
 *  Choice is normal FP code, SSE or 3DNow for Single Precision and normal FP
 *  code or SSE2 for Double Precision.
 * 
 *  ass1XXX - assembly code as original
 *  ass2XXX - assembly code twice original data span
 *  ass4XXX - assembly code four times original data span 
 *
 *  checkNoiseSP() and checkNoiseDP() calculate noise levels to show
 *        accuracy and to show that different versions produce the
 *        same results (approximately particularly SP).
 *
 */
 
#define NAME "Test"
#define TITLE " FFT Graph Benchmark "
#define INITGUID

#define TsizeL2     13  // cache size parameter 13 = 8K FFT 128 KB SP 256
                        // KB DP when segmenting data to fit in L2 cache
                        // for fftSP() and fftDP()
#define TsizeLL     14  // cache size parameter TsizeL2 + 1

#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <direct.h>
#include <commdlg.h>
#include <time.h>
#include <emmintrin.h>
#include <xmmintrin.h>

#include "FFTGraf.h"


//  ************** PROTOPTYPES *****************

BOOL  FAR PASCAL AboutDlgProc( HWND, UINT, WPARAM, LPARAM);
static BOOL InitApp(HINSTANCE, LPSTR, int);
long FAR PASCAL WindowProc(HWND, UINT, WPARAM, LPARAM);                  
static void CleanUpAndPostQuit(void);

BOOL  FAR PASCAL WriteReadGraph( HWND, UINT, WPARAM, LPARAM);
BOOL  FAR PASCAL UpdateCombos( HWND, int, int);

BOOL  reView();
BOOL  drawGraphs(int, int, HDC);

BOOL  saveGrafTxt();
BOOL  saveBmp();
BOOL loadFile();
BOOL compareTxt();
static BOOL GetFileName();

void bitrev(int);
void createExpTableSP(int n, int sign);
void createExpTableDP(int n, int sign);
void createComplexDataDP (int n, double synthFreqI, double synthFreqQ);
void createComplexDataSP (int n, float synthFreqI, float synthFreqQ);

void fftSP(int t, int n); 
void fftDP(int t, int n);
void fftSP2(int t, int n); 
void fftDP2(int t, int n);
void SPCalculate();
void DPCalculate();
void Compare0();
void Compare1();
void Compare2();
void checkNoiseSP();
void checkNoiseDP();
void ass1SP(int, int, int, int);
void ass2SP(int, int, int, int);
void ass4SP(int, int, int, int);
void ass1DP(int, int, int, int);
void ass2DP(int, int, int, int);
void ass4DP(int, int, int, int);
void ass1DPSSE2(int, int, int, int);
void ass2DPSSE2(int, int, int, int);
void ass4DPSSE2(int, int, int, int);
void ass2SPSSE(int, int, int, int);
void ass4SPSSE(int, int, int, int);
void ass1SP3DNow(int, int, int, int);
void ass2SP3DNow(int, int, int, int);
void ass4SP3DNow(int, int, int, int);

// ************* GLOBAL VARIABLES **********

// #define SIZE  1048576 

double   *cDataDP = NULL;   // max 16 MB
double   *cSourceDP;        // max 16 MB
double   *cExpDP;           // max  8 MB
double   *dExpDP;           // max  8 MB
double   *dTable;
double   *cExpDPA;
float    *cDataSP;
float    *cSourceSP;
float    *cExpSP;
float    *dExpSP;
float    *cExpSPA;
int      *table;     // max  4 MB

__m128d *vDP;
__m128d *cDP;


int     RAMmult = 52;
int     RAMunit = 1048576; 
double  ExpReal1DP;
double  ExpImag1DP;
double  ExpReal2DP;
double  ExpImag2DP;
double  ExpReal3DP;
double  ExpImag3DP;
double  ExpReal4DP;
double  ExpImag4DP;
float  ExpReal1SP;
float  ExpImag1SP;
float  ExpReal2SP;
float  ExpImag2SP;
float  ExpReal3SP;
float  ExpImag3SP;
float  ExpReal4SP;
float  ExpImag4SP;

float  PISP;
double PIDP = 3.14159265358979323846;


HWND   mainHWND;

HINSTANCE  MyInstance;
RECT    clientRect;


int     numberFFTs[11];
double  miniSP[11];
double  miniDP[11];
double  miniSPA[11];
double  miniDPA[11];
double  miniSPB[11];
double  miniDPB[11];
int     minFFTRun;
int     FFTsizes = 11;
int     maxFFTRun;
int     FFTpasses = 5;
int     maxruns = 12;
int     maxSecs = 15;

char   currentDir[MAX_PATH];

char   fileMessage[256];

char    logfile[256] = "FFTGraf.txt";
char    bmpfile[256] = "FFTGraf.bmp";

char    commandLine[1024] = "";
char    checkSP[80];
char    checkDP[80];

int  screenWide = 640;
int  screenHigh = 480;
double SPFFT[11][20];
double DPFFT[11][20];
double SP1[11][20];
double SP2[11][20];
double DP1[11][20];
double DP2[11][20];
double SPres;
double DPres;

RECT   recwin;
int    xtraWide; 

char   timeday1[30];
char   grafFile[256];
char   fileToRead[256];
WIN32_FIND_DATA W32DiskData;

int   doRun = 0;
int   tpass = 0;
int   jpass = 0;

int   butT = 365;
int   butL = 380;
int   butW = 80;
int   butH = 22;
int  flash1 = 0;
int  flash2 = 0;
int  runningTest = 0;
int  ramMB;
int  ramKB;
int  vers1 = 0;
int  vers2 = 0;
int  nSP;
int  nDP;
short  keyPress;
char  amsg[100];
char version[50] = " Version 3.00";
char versionU[50];
double maxScale = 0;
double maxVal[40];
int maxScalePos;
char config1[200];
char config2[200];
char config3[200];

BOOL bQuit = FALSE;     
BOOL  loadIt;
BOOL   testRunning = FALSE;
BOOL    displayOnly = FALSE;
BOOL  compareTwo = FALSE;
BOOL   escPressed = FALSE;
BOOL   doSSE2 = FALSE;
BOOL   doSSE1 = FALSE;


// WinMain - initialization, message loop

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                     LPSTR cmdline, int cmdshow)
{
    MSG         msg;
    
    hPrevInstance = hPrevInstance;

    if (!InitApp(hInstance, cmdline, cmdshow))
         return FALSE;
      
    if (maxFFTRun < 10)
    {
        sprintf(amsg, "RAM %dMB, maximum size FFT reduced to %dK",
                          ramMB, numberFFTs[maxFFTRun]);
        MessageBox(NULL, amsg, "Not enough RAM", MB_ICONWARNING | MB_OK);
    }

    //     RAMmult = 52;
    //   max RAMunit = 1048576; 

    cDataDP = (double *)_aligned_malloc(RAMmult * RAMunit, 16);
    if (cDataDP == NULL)
    {
        bQuit = TRUE;
        MessageBox(NULL, "Cannot Allocate Memory", "ERROR", MB_ICONWARNING | MB_OK);
    }
    else
    {
        cSourceDP  = cDataDP + RAMunit*2; // RAMmult 16
        cExpDP     = cDataDP + RAMunit*4; // RAMmult 32
        dExpDP     = cDataDP + RAMunit*5; // RAMmult 40
        dTable     = cDataDP + RAMunit*6; // RAMmult 48
        cDataSP    = (float  *)cDataDP;
        cSourceSP  = (float  *)cSourceDP;
        cExpSP     = (float  *)cExpDP;
        dExpSP     = (float  *)dExpDP;
        table      = (int    *)dTable;
    }

    cExpDPA = (double *)_aligned_malloc(1024, 16);
    if (cExpDPA == NULL)
    {
        bQuit = TRUE;
        MessageBox(NULL, "Cannot Allocate Memory", "ERROR", MB_ICONWARNING | MB_OK);
    }
    cExpSPA = (float *)cExpDPA;

    cDP = (__m128d *)cSourceDP;
    vDP = (__m128d *)cExpDPA;

    while(!bQuit)
    {
        if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
        {
            
           if (msg.message == WM_QUIT)
           {
                CleanUpAndPostQuit();
                break;
           }
   
           if( !GetMessage( &msg, NULL, 0, 0 ) ) return msg.wParam;
           TranslateMessage(&msg);
           DispatchMessage(&msg);
        }
        else if (doRun > 0)
        {
            maxScale = 0;
            if (doRun == 1)
            {
                strcpy(versionU, version);        
                SPCalculate();
                checkNoiseSP();
            }
            else
            {
                DPCalculate();
                checkNoiseDP();
            }
            jpass = jpass + 1;
            if (escPressed)
            {
                jpass = FFTpasses;
                tpass = maxFFTRun + 1;
            }   
            if (jpass == FFTpasses)
            {
                jpass = 0;
                tpass = tpass + 1;
            }
            if (tpass > maxFFTRun - minFFTRun)
            {
                if (doRun == 1 && !escPressed)
                {
                    doRun = 2;
                    flash1 = 0;
                    flash2 = 0;
                    tpass = 0;
                }
                else
                {
                    doRun = 0;
                    if (!escPressed)
                    {
                        saveGrafTxt();
                    }
                    escPressed = FALSE;
                    testRunning = FALSE;
                    flash1 = 0;
                    flash2 = 0;
                    reView();
                }
            }
        }
        else
        {
           // make sure we go to sleep if we have nothing else to do
           WaitMessage();
        }
    }
    if (cDataDP) _aligned_free(cDataDP);
    if (cExpDPA) _aligned_free(cExpDPA);
    return msg.wParam;
    
} // WinMain

//***************************************************************************
//                   Initialization and object creation                     *
//***************************************************************************

// InitApp
// Creates window

static BOOL
InitApp(HINSTANCE hInstance, LPSTR cmdline, int cmdshow)
{
    HWND hwnd;
    WNDCLASS wc;
    LPCTSTR  dir = NULL;
    int i, j, gb;
    char *delims = {" ,"};
//    char commands[20][256];

    char buf[1024] = "";

    
    
    // set up and registers the window class
    
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = sizeof(DWORD);
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon(hInstance, "GenericIcon");
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName = "GenericMenu";
    wc.lpszClassName = NAME;
    if (!RegisterClass(&wc))
        return FALSE;

    
     // Create the window
     

    hwnd =
        CreateWindow(
            NAME,
            TITLE,
            WS_OVERLAPPEDWINDOW, // | WS_HSCROLL | WS_VSCROLL,
            0,
            0,
            screenWide,
            screenHigh,
            NULL, 
            NULL, 
            hInstance, 
            NULL 
        );

    if (!hwnd) return FALSE;

    mainHWND = hwnd;
    strcpy(commandLine, strupr(cmdline));

    WhatConfig();
    ramKB = ramsize / 1024;
    ramMB = ramsize / 1024 / 1024;
    GetCurrentDirectory(MAX_PATH, currentDir);

    if (hasSSE2 && hasSSE2OS) doSSE2 = TRUE;
    if (hasSSE && hasSSEOS)   doSSE1 = TRUE;

    maxFFTRun = 10;     // max used  52 MB FFT 1024 K
    RAMunit = 1048576;
    if (ramMB < 63)
    {
        maxFFTRun = 9;  // max used  26 MB FFT  512 K
        RAMunit = 512 * 1024;
    }
    if (ramMB < 31)
    {
        maxFFTRun = 8;  // max used  13 MB FFT  256 K
        RAMunit = 256 * 1024;
    }
    if (ramMB < 20)
    {
        maxFFTRun = 7;  // max used 6.5 MB FFT  128 K
        RAMunit = 128 * 1024;
    }
    if (ramMB < 11)
    {
        maxFFTRun = 6;  // max used 3.3 MB FFT  64 K
        RAMunit = 64 * 1024;
    }
    minFFTRun = 0;

    gb = 1;
    for (i=0; i<FFTsizes; i++)
    {
        numberFFTs[i] = gb;
        gb = gb * 2;
        for (j=0; j<maxruns; j++)
        {
            SPFFT[i][j] = 0;
            DPFFT[i][j] = 0;
        }
    }
    for (i=0; i<10; i++)
    {
        maxVal[i] = 10.0 - (double)i;
        maxVal[i+9] = (10.0 - (double)i) / 10.0;
        maxVal[i+18] = (10.0 - (double)i) / 100.0;
        maxVal[i+27] = (10.0 - (double)i) / 1000.0;
    }
    strcpy(versionU, version);        
    
    keyPress = GetAsyncKeyState(VK_ESCAPE);

    PISP = (float)PIDP;

    // Display the window
    
    ShowWindow(hwnd, cmdshow);    
    UpdateWindow(hwnd);

    return TRUE;
    
}  // InitApp



// WindowProc - handle messages for the main application window

long FAR PASCAL WindowProc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam)                  
{
    static HINSTANCE hand;
    char   tempFile[256];
    PAINTSTRUCT      ps;
    HDC     hdc;
       
    switch( msg )
    {
    case WM_ACTIVATEAPP:
        break;
        
    case WM_INITMENUPOPUP:
        break;
        
    case WM_CREATE:    
        CreateWindow("BUTTON", "Variables", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                     butL, butT, butW, butH, hwnd, (HMENU)PARAM_GRAPH, hand, NULL);
        CreateWindow("BUTTON", "Exit", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                     butL+butW, butT, butW, butH, hwnd, (HMENU)FINISH_GRAPH, hand, NULL);  
        CreateWindow("BUTTON", "Run", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                     butL+2*butW, butT, butW, butH, hwnd, (HMENU)RUN_BUTTON, hand, NULL);  
        return 0;

    case WM_DESTROY:
        CleanUpAndPostQuit();
        break;

    case WM_COMMAND:
        switch( LOWORD( wparam ) )
        {
        case MENU_ABOUT:
            DialogBox( MyInstance,"AboutBox", hwnd, (DLGPROC)AboutDlgProc );
            break;

        case PARAM_GRAPH:
            DialogBox( MyInstance,"DIALOG_21", hwnd, (DLGPROC)WriteReadGraph );
            reView();                   
            break;

        case MENU_PARAM:
            DialogBox( MyInstance,"DIALOG_21", hwnd, (DLGPROC)WriteReadGraph );
            reView();                   
            break;

        case MENU_SCALE:
            DialogBox( MyInstance,"DIALOG_21", hwnd, (DLGPROC)WriteReadGraph );
            reView();                   
            break;

        case MENU_RUNGRAPH:
            compareTwo = FALSE;
            local_time();
            strcpy(timeday1, timeday);
            WhatConfig();
            ramKB = ramsize / 1024;
            displayOnly = FALSE;
            SetWindowText(mainHWND, " FFT Graph Benchmark  **** RUNNING ****");  
            doRun = 1;
            tpass = 0;
            jpass = 0;
            flash1 = 1;
            flash2 = 0;
            break;
            

        case MENU_LOAD:
            loadIt = TRUE;
            compareTwo = FALSE;
            if (GetFileName())
            {
                FindFirstFile(grafFile, &W32DiskData);
                strcpy (fileToRead, W32DiskData.cFileName);
                loadFile();
                SetCurrentDirectory(currentDir);
            }
            break;

        case MENU_COMPARE:
            loadIt = TRUE;
            compareTwo = FALSE;
            Compare0();
            if (GetFileName())
            {
                FindFirstFile(grafFile, &W32DiskData);
                strcpy (fileToRead, W32DiskData.cFileName);
                loadFile();
                Compare1();    
            }                
            if (GetFileName())
            {
                FindFirstFile(grafFile, &W32DiskData);
                strcpy (fileToRead, W32DiskData.cFileName);
                loadFile();
                Compare2();
                SetCurrentDirectory(currentDir);
            }
            break;
            
        case MENU_SAVE:
            loadIt = FALSE;
            strcpy(tempFile, bmpfile);
            if (GetFileName())
            {
                saveBmp();
                SetCurrentDirectory(currentDir);
                strcpy(bmpfile, tempFile);
            }
            break;

        case MENU_COMPAREOUT:
            if (!compareTwo)
            {
                MessageBox(NULL, "Select results to compare via menu",
                "Not in compare mode", MB_ICONWARNING | MB_OK);
            }
            else
            {
                compareTxt();
            }
            break;

        case RUN_BUTTON:    
            compareTwo = FALSE;
            local_time();
            strcpy(timeday1, timeday);
            WhatConfig();
            ramKB = ramsize / 1024;
            displayOnly = FALSE;
            SetWindowText(mainHWND, " FFT Graph Benchmark  **** RUNNING ****");  
            doRun = 1;
            tpass = 0;
            jpass = 0;
            flash1 = 1;
            flash2 = 0;
            break;
       
        case MENU_VIEWGRAPH:
            reView();
            break;

        case MENU_HELP:
            DialogBox( MyInstance,"DIALOG_44", hwnd, (DLGPROC)AboutDlgProc );                   
            break;
            
        case FINISH_GRAPH:
            CleanUpAndPostQuit();
            break;                

        case MENU_EXIT:
            CleanUpAndPostQuit();
            break;                
        }
        break;
   
        case WM_PAINT:
             hdc = BeginPaint(hwnd, &ps);
             drawGraphs(FFTsizes, FFTsizes, hdc);
             EndPaint(hwnd, &ps);
             break;
           
        case WM_KEYDOWN:
            break;

       case WM_ACTIVATE:
            break;
    } 
    return DefWindowProc( hwnd, msg, wparam, lparam );

} // WindowProc



// AboutDlgProc - processes messages for the about dialog.

BOOL  FAR PASCAL AboutDlgProc( HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam )
{
    lparam = lparam;

    switch( msg ) {
    case WM_INITDIALOG:
        return( TRUE );

    case WM_COMMAND:
        if( LOWORD( wparam ) == IDOK ) {
            EndDialog( hwnd, TRUE );
            return( TRUE );
        }
        break;
    }
    return( FALSE );

} // AboutDlgProc



// CleanUpAndPostQuit

void
CleanUpAndPostQuit(void)
{
    bQuit = TRUE;
    return;
}


BOOL  FAR PASCAL UpdateCombos( HWND hdlg, int minb, int maxb)
{
    minFFTRun = minb;
    maxFFTRun = maxb;

    if (maxScalePos > 0)
    {
        maxScale =  maxVal[maxScalePos - 1];
    }
    else
    {
        maxScale = 0;
    }

    if(minb > maxb)
    {
        maxFFTRun = minb;
    }
    SendDlgItemMessage(hdlg, MINFFTSIZE, CB_SETCURSEL, minFFTRun, 0L);
    SendDlgItemMessage(hdlg, MAXFFTSIZE, CB_SETCURSEL, maxFFTRun, 0L);
    SendDlgItemMessage(hdlg, MAXSCALE, CB_SETCURSEL, maxScalePos, 0L);

    return TRUE;
}

BOOL  FAR PASCAL WriteReadGraph( HWND hdlg, UINT msg,
                                WPARAM wparam, LPARAM lparam )
{
    wparam = wparam;
    lparam = lparam;
    HWND     hwndName;
    int     tempint;
    int     wTextWidth;
    char    temptext[256];
    BOOL     dataOK = TRUE;
    int     i, j, minb, maxb;
    char    wdata[10];
    BOOL    isChanged = FALSE;
    int     tempPasses = FFTpasses;

    minb = minFFTRun;
    maxb = maxFFTRun;
    
    if (maxScale == 0) maxScalePos = 0;
    
    switch( msg )
     {
       case WM_INITDIALOG:
        SetDlgItemInt(hdlg, FFT_PASSES,  FFTpasses, FALSE); 
        SetDlgItemInt(hdlg, MAXRUNSECS,  maxSecs, FALSE); 

        for (i=0; i<FFTsizes; i++)
        {
           sprintf(wdata,"%4d", numberFFTs[i]);
           SendDlgItemMessage(hdlg, MINFFTSIZE, CB_ADDSTRING, 0, (LPARAM) wdata);
           SendDlgItemMessage(hdlg, MAXFFTSIZE, CB_ADDSTRING, 0, (LPARAM) wdata);
        }
        sprintf(wdata,"N/A");
        SendDlgItemMessage(hdlg, MAXSCALE, CB_ADDSTRING, 0, (LPARAM) wdata);
        for (i=0; i<37; i++)
        {
            sprintf(wdata,"%6.3f", maxVal[i]);
            SendDlgItemMessage(hdlg, MAXSCALE, CB_ADDSTRING, 0, (LPARAM) wdata);
        }
        UpdateCombos(hdlg, minb, maxb);


        hwndName = GetDlgItem(hdlg, GRAPH_FILE);
        SendMessage(hwndName, WM_SETTEXT, 0,
                                    (LPARAM) ((LPSTR) logfile));

        return( TRUE );

      case WM_COMMAND:
        if( LOWORD( wparam ) ==  DO_RUN)
        {
             hwndName = GetDlgItem(hdlg, GRAPH_FILE);
             wTextWidth = SendMessage(hwndName, WM_GETTEXT,
                      sizeof(temptext), (LPARAM) ((LPSTR) temptext));
             if (wTextWidth > 0)
             {
                  lstrcpy(logfile, temptext);
             }
             else
             {
                 dataOK = FALSE;
             }
             tempint = GetDlgItemInt(hdlg, FFT_PASSES, NULL, FALSE);
             if (tempint > 0 && tempint < maxruns+1)
             {
                 FFTpasses = tempint;
                 if (tempPasses != FFTpasses) isChanged = TRUE;
             }
             else
             {
                 dataOK = FALSE;
             }
             tempint = GetDlgItemInt(hdlg, MAXRUNSECS, NULL, FALSE);
             if (tempint > 0)
             {
                 maxSecs = tempint;
             }
             else
             {
                 dataOK = FALSE;
             }        
             if (dataOK)
             {
                if (isChanged)
                {
                    for (i=0; i<FFTsizes; i++)
                    {
                        for (j=0; j<maxruns; j++)
                        {
                            SPFFT[i][j] = 0;
                            DPFFT[i][j] = 0;
                        }
                    }
                }
                EndDialog( hdlg, TRUE );
                return (TRUE);
             }
             else
             {
                MessageBox(NULL, " Parameters wrong. Try again", "Error", MB_ICONWARNING | MB_OK);
                EndDialog( hdlg, TRUE );
                return (FALSE);
             }                 
        }
        else if( LOWORD( wparam ) == IDCANCEL )
        {
            if (isChanged)
            {
                for (i=0; i<FFTsizes; i++)
                {
                    for (j=0; j<maxruns; j++)
                    {
                        SPFFT[i][j] = 0;
                        DPFFT[i][j] = 0;
                    }
                }
            }
            EndDialog( hdlg, TRUE );    
            return (FALSE);
        }
        else if( HIWORD( wparam ) == CBN_SELCHANGE)
        {
            switch (LOWORD (wparam))
            {
                case MINFFTSIZE:
                  minb = (int)SendDlgItemMessage(hdlg, MINFFTSIZE, CB_GETCURSEL, 0, 0);
                  if (minb != minFFTRun) isChanged = TRUE;  
                  break;
        
                case MAXFFTSIZE:
                  maxb = (int)SendDlgItemMessage(hdlg, MAXFFTSIZE, CB_GETCURSEL, 0, 0);
                  if (maxb != maxFFTRun) isChanged = TRUE;
                  break;

                case MAXSCALE:
                  maxScalePos = (int)SendDlgItemMessage(hdlg, MAXSCALE, CB_GETCURSEL, 0, 0);
                  break;
            }
            if (isChanged)
            {
                for (i=0; i<FFTsizes; i++)
                {
                    for (j=0; j<maxruns; j++)
                    {
                        SPFFT[i][j] = 0;
                        DPFFT[i][j] = 0;
                    }
                }
            }
            UpdateCombos(hdlg, minb, maxb);
        }
     }
     return( FALSE );

} // WriteReadGraph


BOOL  drawGraphs(int nw, int nr, HDC hdc)
{
    HBRUSH hbr;
    HPEN hpen;

    int  thissize;
    char thisdata[200];
    TEXTMETRIC tmt;
    int     charHeight;
    int     charWidth;
    PLOGFONT  plf;
    HFONT   hfont;
    int   nPtSize;   
    int   text2 =  5;   
    int   yTop1 =    44;  // 20
    int   yBottom1 = 330; // 340
    int   yInc1;
    double   scale1;
    double   mMax;
    double   sMax;
    int   height1;
    int   wide1;
    int   xStart = 135;  // 140;
    int   xEnd = 615; // 620
    int   yBottom2;
    int   yTitle1 = 400;
    int   i, j, m, g;
    int   pixelsOn;
    int   incH;
    int   incW;
    char  sSP[10];
    char  sDP[10];
    double minSPc;
    double minDPc;
    double SPFFTU[11][20];
    double DPFFTU[11][20];
    BOOL    higher;


    yBottom2 = butT;

    height1 = yBottom1 - yTop1;
    incH = height1 / 11;
    yInc1 = height1/10;
    wide1 = xEnd - xStart;
    incW = wide1 / 10;

    GetClientRect(mainHWND, &recwin);
    j = (recwin.right + 3) / 4;
    xtraWide = j * 4  - recwin.right;
    if (xtraWide == 1) recwin.right = recwin.right - 3;
    if (xtraWide == 2) recwin.right = recwin.right - 2;
    if (xtraWide == 3) recwin.right = recwin.right - 1;


    // Max Scale
    for (i=0; i<FFTsizes; i++)
    {
        for (j=0; j<maxruns; j++)
        {
            SPFFTU[i][j] = SPFFT[i][j];
            DPFFTU[i][j] = DPFFT[i][j];
            if (maxScale > 0)
            {
                if (SPFFTU[i][j] > maxScale) SPFFTU[i][j] = maxScale;
                if (DPFFTU[i][j] > maxScale) DPFFTU[i][j] = maxScale;
            }
        }
    }

    // Max times and average millisecs/K

    for (i=0; i<FFTsizes; i++)
    {
        miniSP[i] = 0;
        miniDP[i] = 0;
    }        
    mMax = 0;
    for (i=minFFTRun; i<maxFFTRun+1; i++)
    {
        minSPc = 10000000;
        minDPc = 10000000;
        for (j=0; j<FFTpasses; j++)
        {
            if (SPFFTU[i][j] > mMax) mMax = SPFFTU[i][j];
            if (DPFFTU[i][j] > mMax) mMax = DPFFTU[i][j];
            if (SPFFT[i][j] > 0 && SPFFT[i][j] < minSPc) minSPc = SPFFT[i][j]; 
            if (DPFFT[i][j] > 0 && DPFFT[i][j] < minDPc) minDPc = DPFFT[i][j]; 
        }
        if (minSPc < 10000000) miniSP[i] = minSPc * (double)numberFFTs[i];
        if (minDPc < 10000000) miniDP[i] = minDPc * (double)numberFFTs[i];
    }
    if (maxScale > 0) mMax = maxScale;

    // adjust scale

    sMax = 0.001;
    for (j=1; j<11; j++)
    {
        for (i=1; i<11; i++)
        {
            if (mMax < sMax * (double)i || mMax == sMax * (double)i)
            {
               scale1 = sMax * (double)i;
               i = 11;
               j = 11;
            }
        }
        sMax = sMax * 10;
    }


    // Fill most of screen white

    hbr = CreateSolidBrush(RGB(255,255,255));  
    SelectObject(hdc, hbr);
    hpen = CreatePen(PS_SOLID, 0, RGB(255,255,255));
    SelectObject(hdc, hpen);
    Rectangle(hdc, recwin.left, recwin.top, recwin.right, butT - 1);
    Rectangle(hdc, recwin.left, butT-1, butL - 1, butT+butH+1);
    Rectangle(hdc, recwin.left, butT+butH+1, recwin.right, recwin.bottom);
    DeleteObject(hpen);
    
    // Text black on white

    SetBkColor(hdc, RGB(255,255,255));
    SetTextColor(hdc, RGB(0,0,0)); 

    // Heading Font
    
    plf = (PLOGFONT) LocalAlloc(LPTR, sizeof(LOGFONT));
    nPtSize = 12; // 10
    plf->lfHeight = -MulDiv(nPtSize, GetDeviceCaps(hdc, LOGPIXELSY), 72);
    strcpy(plf->lfFaceName, "Times New Roman");
    hfont = CreateFontIndirect(plf);
    SelectObject(hdc, hfont);

    GetTextMetrics(hdc, &tmt);
    charHeight = tmt.tmHeight;
    charWidth  = tmt.tmAveCharWidth;

    if (testRunning)
    {
        SetTextColor(hdc, RGB(255,0,0)); 
        sprintf(thisdata, "Test Running - Press Esc Key To Kill If Running Time Too Long                  ");
        thissize =  strlen(thisdata);
        ExtTextOut(hdc, xStart+80, 3, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
    }
    else
    {
        if (!compareTwo)
        {
            sprintf(thisdata, "Scott's & Roy's FFTGraf Benchmark %s                                      ", versionU);
            thissize =  strlen(thisdata);
            ExtTextOut(hdc, xStart+80, 3, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        }
        else
        {
            if (vers1 ==  vers2)
            {
                sprintf(thisdata, "FFT Results Comparison Version %d Tests                               ", vers1);
            }
            else
            {
                SetTextColor(hdc, RGB(255,0,0)); 
                sprintf(thisdata, "Comparison WARNING A = Version %d B = %d                              ", vers1, vers2);
            }
            thissize =  strlen(thisdata);
            ExtTextOut(hdc, xStart+80, 3, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        }             
    }
    SetTextColor(hdc, RGB(0,0,0)); 

    LocalFree((HLOCAL) plf);
    DeleteObject(hfont);

    // Font
    
    plf = (PLOGFONT) LocalAlloc(LPTR, sizeof(LOGFONT));
    nPtSize = 8;
    plf->lfHeight = -MulDiv(nPtSize, GetDeviceCaps(hdc, LOGPIXELSY), 72);
    strcpy(plf->lfFaceName, "Courier New");
    hfont = CreateFontIndirect(plf);
    SelectObject(hdc, hfont);

    GetTextMetrics(hdc, &tmt);
    charHeight = tmt.tmHeight;
    charWidth  = tmt.tmAveCharWidth;

    hpen = CreatePen(PS_SOLID, 0, RGB(0,0,0));
    SelectObject(hdc, hpen);
     
    // Headings

    if (!compareTwo)
    {
        sprintf(thisdata, "Memory used 16-28 x SP size, 28-52 x DP size, maximum 52 MB"); 
        thissize =  strlen(thisdata);
        ExtTextOut(hdc, xStart+50, charHeight+10, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        sprintf(thisdata, "Size   Min Millisecs  ");
    }
    else
    {
        sprintf(thisdata, "  Speed Comparison of running times  >1 A is faster      ");
        thissize =  strlen(thisdata);
        ExtTextOut(hdc, xStart+50, charHeight+10, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        sprintf(thisdata, "Size   KBytes Used  ");
    }
    thissize =  strlen(thisdata);
    ExtTextOut(hdc, 10, charHeight-2, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
    sprintf(thisdata, " K     SP     DP     ");
    thissize =  strlen(thisdata);
    ExtTextOut(hdc, 10, charHeight*2-2, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);


    // X titles

    if (!compareTwo)
    {
        sprintf(thisdata, "Millisecs / K Size  ");
    }
    else
    {
        sprintf(thisdata, "Millisecs Ratio B/A ");
    }
    thissize = strlen(thisdata);;
    ExtTextOut(hdc, xStart + 150, yBottom1+charHeight+1, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);


    // Variables

    if (!compareTwo)
    {
        sprintf(thisdata, " %d Passes, Min %dK, Max %dK, Max Seconds %d    ", FFTpasses,
                                  numberFFTs[minFFTRun], numberFFTs[maxFFTRun], maxSecs);
        thissize = strlen(thisdata);
        ExtTextOut(hdc, 15, yBottom2, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
    }
    else
    {
        sprintf(thisdata, "A %s  ", config1);
        thissize = strlen(thisdata);
        ExtTextOut(hdc, 15, yBottom2, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
    }
    if (!compareTwo)
    {
       sprintf(thisdata, " Memory Size %d KBytes     ", ramKB);
    }
    else
    {
       sprintf(thisdata, "B %s  ", config2);
    }                                     
    thissize = strlen(thisdata);
    ExtTextOut(hdc, 15, yBottom2+charHeight, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
     

    // Colour keys

    sprintf(thisdata, "Key ");
    thissize = 4;
    ExtTextOut(hdc, 5, yBottom1+5, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);

    SetBkColor(hdc, RGB(255,0,0));
    SetTextColor(hdc, RGB(255,255,255)); 
    sprintf(thisdata, "  SP");
    thissize = 4;
    ExtTextOut(hdc, 45, yBottom1+5, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);

    SetBkColor(hdc, RGB(0,0,255));
    sprintf(thisdata, "  DP");
    thissize = 4;
    ExtTextOut(hdc, 95, yBottom1+5, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);

    SetTextColor(hdc, RGB(0,0,0)); 


    // Graph Box
    
    MoveToEx(hdc, xStart, yTop1, NULL);
    LineTo(hdc, xStart, yBottom1);
    LineTo(hdc, xEnd, yBottom1);
    LineTo(hdc, xEnd, yTop1);
    LineTo(hdc, xStart-3, yTop1);

    SetBkColor(hdc, RGB(255,255,255));

    // Size & results

    j = 0;
    for (i=yTop1-2; i<yBottom1-2; i=i+incH)
    {
        sprintf(thisdata, "%4d", numberFFTs[j]);
        if (!compareTwo)
        {
            SPres = miniSP[j];
            sprintf(sSP, "%7.0f", SPres);
            if (SPres < 100) sprintf(sSP, "%7.1f", SPres);
            if (SPres < 10) sprintf(sSP, "%7.2f", SPres);
            if (SPres < 1) sprintf(sSP, "%7.3f", SPres);
            if (SPres == 0) sprintf(sSP, "       ");
        }
        else
        {
            SPres = 16.0 * (double)numberFFTs[j];
            sprintf(sSP, "%7.0f", SPres);
            if (j > 3) strcat(sSP, "+");
        }

        if (!compareTwo)
        {
            DPres = miniDP[j];
            sprintf(sDP, "%7.0f", DPres);
            if (DPres < 100) sprintf(sDP, "%7.1f", DPres);
            if (DPres < 10) sprintf(sDP, "%7.2f", DPres);
            if (DPres < 1) sprintf(sDP, "%7.3f", DPres);
            if (DPres == 0) sprintf(sDP, "       ");
        }
        else
        {
            DPres = 28.0 * (double)numberFFTs[j];
            sprintf(sDP, "%7.0f", DPres);
             if (j > 3) strcat(sDP, "+");
       }

        if (flash1 == 1 && runningTest == j)
        {
            SetBkColor(hdc, RGB(255,0,0));
            SetTextColor(hdc, RGB(255,255,255)); 
        }
        else if (flash1 == 2 && runningTest == j)
        {
            SetBkColor(hdc, RGB(0,255,255));
            SetTextColor(hdc, RGB(255,0,0)); 
        }
        else
        {
            SetBkColor(hdc, RGB(255,255,255));        
            SetTextColor(hdc, RGB(0,0,0)); 
        }  
        strcat(thisdata, sSP);        
        thissize = strlen(thisdata);
        ExtTextOut(hdc, 0, i, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        SetBkColor(hdc, RGB(255,255,255));        
        SetTextColor(hdc, RGB(0,0,0)); 
        if (flash2 == 1 && runningTest == j)
        {
            SetBkColor(hdc, RGB(0,0,255));
            SetTextColor(hdc, RGB(255,255,255)); 
        }
        else if (flash2 == 2 && runningTest == j)
        {
            SetBkColor(hdc, RGB(0,255,255));
            SetTextColor(hdc, RGB(255,0,0)); 
        }
        else
        {
            SetBkColor(hdc, RGB(255,255,255));        
            SetTextColor(hdc, RGB(0,0,0)); 
        }
        sprintf(thisdata, "           "); // - 1 space
        strcat(thisdata, sDP);        
        thissize = strlen(thisdata);
        ExtTextOut(hdc, 0, i+incH/2, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        SetBkColor(hdc, RGB(255,255,255));        
        SetTextColor(hdc, RGB(0,0,0)); 
        j = j + 1;
    }

    // Scale

    sprintf(thisdata, "0");
    thissize = strlen(thisdata);;
    ExtTextOut(hdc, xStart-charWidth / 2, yBottom1+1, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
    
    j = 1;
    for (i=xStart; i<xEnd-incW; i=i+incW)
    {
        m = 0;
        sprintf(thisdata, "%d", (int)(scale1 * j / 10));
        if (scale1 < 10) sprintf(thisdata, "%3.1f", (scale1 * j) / 10);
        if (scale1 < 1) sprintf(thisdata, "%4.2f", (scale1 * j) / 10);
        if (scale1 < 0.1) sprintf(thisdata, "%5.3f", (scale1 * j) / 10);
        if (scale1 < 0.01) sprintf(thisdata, "%6.4f", (scale1 * j) / 10);
        thissize = strlen(thisdata);
        if (thissize == 1) m = 1;
        if (thissize > 2) m = -1;
        ExtTextOut(hdc, i+incW- charWidth * (thissize - 1) - m * charWidth / 2,
           yBottom1+1, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
        j = j + 1;
    }

    m = 0;
    sprintf(thisdata, "%d", (int)scale1); 
    if (scale1 < 10) sprintf(thisdata, "%3.1f", scale1);
    if (scale1 < 1) sprintf(thisdata, "%4.2f", scale1);
    if (scale1 < 0.1) sprintf(thisdata, "%5.3f", scale1);
    if (scale1 < 0.01) sprintf(thisdata, "%6.4f", scale1);
    thissize = strlen(thisdata);
    if (thissize == 1) m = 1;
    ExtTextOut(hdc, xEnd - charWidth * (thissize - 1) - m * charWidth / 2,
      yBottom1+1, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);

    // Grid lines

    for (i=1; i<10; i++)
    {
       MoveToEx(hdc, xStart + i * incW, yBottom1+1, NULL);
       LineTo(hdc, xStart + i * incW, yTop1);
    }


    DeleteObject(hbr);

    DeleteObject(hpen);

    // SP graphs Red
    
    hpen = CreatePen(PS_SOLID, 0, RGB(255,0,0));
    SelectObject(hdc, hpen);
    g = 1; // 4

    for (i=0; i<nw; i++)
    {
        higher = FALSE;
        for (j=0; j<maxruns; j++)
        {
            pixelsOn = (int)((double)wide1 / scale1 * SPFFTU[i][j]);
            MoveToEx(hdc, xStart, yTop1+i*incH+g+j, NULL);
            LineTo(hdc, xStart+pixelsOn, yTop1+i*incH+g+j);
            if (SPFFT[i][j] > SPFFTU[i][j]) higher = TRUE;
        }
        if (higher)
        {
            SetTextColor(hdc, RGB(255,0,0)); 
            ExtTextOut(hdc, xEnd, yTop1+incH*i-3, ETO_CLIPPED, &recwin, ">", 1, (LPINT) NULL);
            SetTextColor(hdc, RGB(0,0,0));
        }  
    }
    DeleteObject(hpen);

    // DP graphs Blue
    
    hpen = CreatePen(PS_SOLID, 0, RGB(0,0,255));
    SelectObject(hdc, hpen);
    g = 14; // 17
    for (i=0; i<nr; i++)
    {
        higher = FALSE;
        for (j=0; j<maxruns; j++)
        {
            pixelsOn = (int)((double)wide1 / scale1 * DPFFTU[i][j]);
            MoveToEx(hdc, xStart, yTop1+i*incH+g+j, NULL);
            LineTo(hdc, xStart+pixelsOn, yTop1+i*incH+g+j);
            if (DPFFT[i][j] > DPFFTU[i][j]) higher = TRUE;
        }
        if (higher)
        {
            SetTextColor(hdc, RGB(0,0,255)); 
            ExtTextOut(hdc, xEnd, yTop1+incH*i+incH/2-3, ETO_CLIPPED, &recwin, ">", 1, (LPINT) NULL);
            SetTextColor(hdc, RGB(0,0,0));
        }  
    }
    DeleteObject(hpen);

    hpen = CreatePen(PS_SOLID, 0, RGB(255,255,255));
    SelectObject(hdc, hpen);


    // Date and Configuration - bottom

    if (!compareTwo)
    {
        sprintf(thisdata, "%s ", configData1);
    }
    else
    {
        sprintf(thisdata, "A %s ", configData1);
    }        
    if (!displayOnly)
    {
        local_time();
    }
    if (!compareTwo)
    {
        strncat(thisdata, timeday, strlen(timeday)-1);
    }
    thissize = strlen(thisdata);
    ExtTextOut(hdc, 15, yBottom2 + charHeight*2, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);
    
    if (!compareTwo)
    {
        sprintf(thisdata, "%s ", configData2);
    }
    else
    {
        sprintf(thisdata, "B %s ", configData2);
    }
    thissize = strlen(thisdata);
    ExtTextOut(hdc, 15, yBottom2 + charHeight*3, ETO_CLIPPED, &recwin, thisdata, thissize, (LPINT) NULL);

    DeleteObject(hpen);

    LocalFree((HLOCAL) plf);
    DeleteObject(hfont);

    ReleaseDC(mainHWND, hdc); 
    return TRUE;

} // drawGraphs


BOOL reView()
{
    HDC   hdc;
    
    SetWindowText(mainHWND, " FFT Graph Benchmark  ");  
    hdc = GetDC(mainHWND);
    drawGraphs(FFTsizes, FFTsizes, hdc);
    ReleaseDC(mainHWND, hdc); 
    return TRUE;
}


BOOL saveGrafTxt()
{
     int i, j;
     char  sSP[10];
     char  sDP[10];
     FILE *outfile;

     SetCurrentDirectory(currentDir);
     outfile = fopen(logfile,"a+");
     if (outfile == NULL)
     {
         MessageBox(NULL, "Cannot open log file", logfile,
                                 MB_ICONWARNING | MB_OK);
         return FALSE;
     }
     fprintf (outfile, "\n\n ###########################################################\n");
     fprintf (outfile, "     FFTGraf Test %s %s \n", version, timeday1);
     fprintf(outfile, " By Roy Longbottom via Scott Taylor's code and now SSE, SSE2, 3DNow\n\n");
     fprintf(outfile, " %s\n", configData1);
     fprintf(outfile, " %s\n", configData2);
     fprintf(outfile, " %s\n", configData3);
     fprintf(outfile, "  %d Passes Min %d K Max %d K Max Seconds %d\n\n", FFTpasses,
                       numberFFTs[minFFTRun], numberFFTs[maxFFTRun], maxSecs);
   
     fprintf (outfile, " Size  Single Precision FFTs ");
     if (doSSE1)
     {
        fprintf (outfile, "using SSE\n");
     }
     else if (has3DNow)
     {
        fprintf (outfile, "using 3DNow\n");
     }
     else
     {
        fprintf (outfile, "\n");
     }
     fprintf (outfile, "    K  Millisecond each pass\n"); 

     for (i=minFFTRun; i<maxFFTRun+1; i++) 
     {
        fprintf (outfile, "%5d", numberFFTs[i]);
        for (j=0; j<FFTpasses; j++)
        {
            SPres = SPFFT[i][j] * (double)numberFFTs[i];
            sprintf(sSP, "%7.0f", SPres);
            if (SPres < 100) sprintf(sSP, "%7.1f", SPres);
            if (SPres < 10) sprintf(sSP, "%7.2f", SPres);
            if (SPres < 1) sprintf(sSP, "%7.3f", SPres);
            if (SPres == 0) sprintf(sSP, "       ");
            fprintf (outfile, "%s", sSP);
        }
        fprintf (outfile, "\n");
     }
     fprintf (outfile, "\n Size  Double Precision FFTs ");
     if (doSSE2)
     {
        fprintf (outfile, "using SSE2\n");
     }
     else
     {
        fprintf (outfile, "\n");
     }
     fprintf (outfile, "    K  Millisecond each pass\n"); 
    
     for (i=minFFTRun; i<maxFFTRun+1; i++)
     {
        fprintf (outfile, "%5d", numberFFTs[i]);
        for (j=0; j<FFTpasses; j++)
        {
            DPres = DPFFT[i][j] * (double)numberFFTs[i];
            sprintf(sDP, "%7.0f", DPres);
            if (DPres < 100) sprintf(sDP, "%7.1f", DPres);
            if (DPres < 10) sprintf(sDP, "%7.2f", DPres);
            if (DPres < 1) sprintf(sDP, "%7.3f", DPres);
            if (DPres == 0) sprintf(sDP, "       ");
            fprintf (outfile, "%s", sDP);
        }
        fprintf (outfile, "\n");
     }
     if (maxFFTRun == 10)
     {
         fprintf (outfile, "\n%s\n", checkSP);
         fprintf (outfile, "%s\n", checkDP);
     }
     local_time();
     fprintf (outfile, "\n               End FFT Test %s\n", timeday);
     fflush (outfile);
     fclose(outfile);
     return TRUE;

} // saveGrafTxt 


BOOL  saveBmp()
{

    HANDLE            hFile;
    LPBITMAPINFO      lpbmi = NULL;
    BITMAPFILEHEADER  bfh;
    int              dataLength;
    HDC               hdc = NULL;
    int              fileLength;
    DWORD             BytesWritten;
    char             *outBytes;
    BYTE              b;
    int              i, j, n;
    DWORD             g1, g2;
    HDC     hdcImage = NULL;
    HBITMAP hbmpImage = NULL;
    BITMAP  bmpImage;

    PALETTEENTRY Vga16Palette[16] =
    {
        { 0,   0,   0,    0 },
        { 0x80,0,   0,    0 },
        { 0,   0x80,0,    0 },
        { 0x80,0x80,0,    0 },
        { 0,   0,   0x80, 0 },
        { 0x80,0,   0x80, 0 },
        { 0,   0x80,0x80, 0 },
        { 0x80,0x80,0x80, 0 },
        { 0xC0,0xC0,0xC0, 0 },
        { 0xFF,0,   0,    0 },
        { 0,   0xFF,0,    0 },
        { 0xFF,0xFF,0,    0 },
        { 0,   0,   0xFF, 0 },
        { 0xFF,0,   0xFF, 0 },
        { 0,   0xFF,0xFF, 0 },
        { 0xFF,0xFF,0xFF, 0 }
    };
    hdc = GetDC(mainHWND);

    // Allocate and fill out a BITMAPINFO struct for the new DIB

    dataLength = (recwin.right * recwin.bottom) / 2;

    outBytes = (char *) malloc(dataLength);
    if (outBytes == NULL)
    {
        MessageBox(mainHWND, "Cannot allocate memory",
               bmpfile, MB_ICONINFORMATION | MB_OK);
        ReleaseDC(mainHWND, hdc);
        return FALSE;
    }

    UpdateWindow(mainHWND);    

    hbmpImage = CreateCompatibleBitmap(hdc, recwin.right, recwin.bottom);
    GetObject(hbmpImage, sizeof(BITMAP), &bmpImage);
    hdcImage = CreateCompatibleDC(hdc);
    SelectObject(hdcImage, hbmpImage);
    BitBlt(hdcImage, 0, 0, recwin.right, recwin.bottom, hdc, 0, 0, SRCCOPY);
    n = 0;
    for (i=0; i<recwin.bottom; i++)
    {
        for (j=0; j<recwin.right; j=j+2)
        {
            b = 0x00;
            g1 = GetPixel(hdcImage, j+0, recwin.bottom-1-i);
            g2 = GetPixel(hdcImage, j+1, recwin.bottom-1-i);
            
            if(g1 == RGB(255,0,0)) b = b | 0xC0;
            if(g2 == RGB(255,0,0)) b = b | 0x0C;
            if(g1 == RGB(0,0,255)) b = b | 0x90;
            if(g2 == RGB(0,0,255)) b = b | 0x09;
            if(g1 == RGB(255,255,255)) b = b | 0xF0;
            if(g2 == RGB(255,255,255)) b = b | 0x0F;
            outBytes[n] = b;
            n=n+1;
        }
    }
    DeleteDC(hdcImage);
    DeleteObject(hbmpImage);

    lpbmi = (BITMAPINFO*) malloc( sizeof( BITMAPINFO ) + dataLength + 264 );
    if (lpbmi == NULL)
    {
        MessageBox(mainHWND, "Cannot allocate memory",
               bmpfile, MB_ICONINFORMATION | MB_OK);
        ReleaseDC(mainHWND, hdc);
        free(outBytes);
        return FALSE;
    }
    lpbmi->bmiHeader.biSize = sizeof( BITMAPINFOHEADER );
    lpbmi->bmiHeader.biWidth = recwin.right;
    lpbmi->bmiHeader.biHeight = recwin.bottom;
    lpbmi->bmiHeader.biPlanes = 1;
    lpbmi->bmiHeader.biBitCount = 4;
    lpbmi->bmiHeader.biCompression = BI_RGB;
    lpbmi->bmiHeader.biSizeImage = dataLength;
    lpbmi->bmiHeader.biXPelsPerMeter = 0;
    lpbmi->bmiHeader.biYPelsPerMeter = 0;
    lpbmi->bmiHeader.biClrUsed = 0;
    lpbmi->bmiHeader.biClrImportant = 0;

    memcpy(lpbmi->bmiColors, Vga16Palette, 64);    
    memcpy(lpbmi->bmiColors+64, outBytes, dataLength);    
    fileLength = 54 + 64 + dataLength;

    hFile = CreateFile(bmpfile, GENERIC_WRITE, 0, NULL,
                        CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {
         MessageBox(mainHWND, "Cannot open file", bmpfile,
                         MB_ICONINFORMATION | MB_OK);
        ReleaseDC(mainHWND, hdc);
        free( lpbmi );
        free(outBytes);
        return FALSE;
    }
    
    // BMP file header

    bfh.bfType = 0x4d42;
    bfh.bfSize = fileLength;
    bfh.bfReserved1 = 0;
    bfh.bfReserved2 = 0;
    bfh.bfOffBits =  118;

    WriteFile( hFile, &bfh, 14, &BytesWritten, NULL );
    WriteFile( hFile, lpbmi, 104, &BytesWritten, NULL );
    WriteFile( hFile,  lpbmi->bmiColors+64, dataLength, &BytesWritten, NULL ); 

    CloseHandle( hFile ); 

    free( lpbmi );
    free(outBytes);
    ReleaseDC(mainHWND, hdc);
    
    return TRUE;
    
} // saveBmp()
    

static BOOL GetFileName()
{
    static char    filterList1[] = "File (*.txt)" \
                                        "\0" \
                                        "*.txt" \
                                        "\0\0";
    static char    filterList2[] = "File (*.bmp)" \
                                        "\0" \
                                        "*.bmp" \
                                        "\0\0";
    OPENFILENAME        of;
    int                 rc;
    
    if (loadIt)
    {
        grafFile[ 0 ] = 0;
    }
    else
    {
        strcpy(grafFile, bmpfile);
    }
    memset( &of, 0, sizeof( OPENFILENAME ) );
    of.lStructSize = sizeof( OPENFILENAME );
    of.hwndOwner = mainHWND;
    of.lpstrFilter = (LPSTR) filterList1;
    if (!loadIt) of.lpstrFilter = (LPSTR) filterList2;
    of.lpstrDefExt = "";
    of.nFilterIndex = 1L;
    of.lpstrFile = grafFile;
    of.nMaxFile = _MAX_PATH;
    of.lpstrTitle = NULL;
    of.Flags = OFN_HIDEREADONLY;
    of.lpstrInitialDir = currentDir;
    
    if (loadIt)
    {
        rc = GetOpenFileName( &of );
    }
    else
    {
        rc = GetSaveFileName( &of );
        strcpy(bmpfile, of.lpstrFile);
    }
   
    return( rc );
    
} // GetFileName

BOOL loadFile()
{
     HDC   hdc;
     FILE *infile;
     char buf[350];
     char buf2[350];
     char buf3[350];
     BOOL fin = FALSE;
     BOOL skip;
     int  gotGraf = 0;
     int  i, j, i1;
     char fpnumber[10];
     char v[10];
     int  mbYNC;
     int  minKB, maxKB;
     int  grafCount = 0;
     int  count2 = 0;

     infile = fopen(fileToRead,"r");
     if (infile == NULL)
     {
         MessageBox(NULL, "Cannot open graph log file", logfile,
                                 MB_ICONWARNING | MB_OK);
         return FALSE;
     }

    while (!fin && fgets(buf, 120, infile) != NULL)
    {
        if (strnicmp(buf, "     FFTGraf Test", 17) == 0)
        {
            grafCount = grafCount + 1;
        }
     }
     fclose(infile);

     infile = fopen(fileToRead,"r");
     if (infile == NULL)
     {
         MessageBox(NULL, "Cannot open graph log file", logfile,
                                 MB_ICONWARNING | MB_OK);
         return FALSE;
     }

    while (!fin && fgets(buf, 120, infile) != NULL)
    {
        skip = FALSE;
        if (strnicmp(buf, "     FFTGraf Test", 17) == 0)
        {
            for (i=0; i<FFTsizes; i++)
            {
                for (j=0; j<maxruns; j++)
                {
                    SPFFT[i][j] = 0;
                    DPFFT[i][j]  = 0;
                }
            }
            sscanf(buf, "%27c%s%25c ", buf2, v, timeday);
            sprintf(versionU, " Version %s", v);
            strcat(timeday, " ");
            vers2 = (int)strtod(v, NULL);
            count2 = count2 + 1;

            fgets(buf, 120, infile);
            fgets(buf, 120, infile);
            fgets(buf, 120, infile);
            fgets(buf, 2, infile);
            fgets(buf, 120, infile);
            strncpy(configData1, buf, strlen(buf) - 1);
            configData1[strlen(buf)-1] = 0;
            fgets(buf, 2, infile);
            fgets(buf, 120, infile);
            strncpy(configData2, buf, strlen(buf) - 1);
            configData2[strlen(buf)-1] = 0;
            fgets(buf, 1, infile);
            fgets(buf, 120, infile);
            strncpy(configData3, buf, strlen(buf) - 1);
            sscanf(buf, "%s %d", buf, &ramKB);
            configData3[strlen(buf)] = 0;
            displayOnly = TRUE;
            fscanf(infile, "%d %s %s %d %s %s %d %s %s %s %d",
                &FFTpasses, buf, buf, &minKB, buf, buf, &maxKB, buf, buf, buf, &maxSecs);
            for (i=0; i<FFTsizes; i++)
            {
                if (numberFFTs[i] == minKB) minFFTRun = i;
                if (numberFFTs[i] == maxKB) maxFFTRun = i;
            }
            fgets(buf, 120, infile);
            fgets(buf, 120, infile);
            fgets(buf, 120, infile);
            if (strnicmp(buf, " Size", 5) != 0) skip = TRUE;
            if (!skip)
            {
                fgets(buf, 120, infile);
                for (i=minFFTRun; i<maxFFTRun+1; i++)
                {
                    fscanf(infile, "%5d", &i1);
                    for (j=0; j<FFTpasses; j++)
                    {
                        fscanf(infile, "%7c", fpnumber);
                        SPFFT[i][j] = strtod(fpnumber, NULL) / (double)numberFFTs[i];
                    }
                    fgets(buf, 120, infile);  // CR
                }
                fgets(buf, 120, infile);
                fgets(buf, 120, infile);
                if (strnicmp(buf, " Size", 5) != 0) skip = TRUE;
            }
            if (!skip)
            {                
                fgets(buf, 120, infile);
                for (i=minFFTRun; i<maxFFTRun+1; i++)
                {
                    fscanf(infile, "%5d", &i1);
                    for (j=0; j<FFTpasses; j++)
                    {
                        fscanf(infile, "%7c", fpnumber);
                        DPFFT[i][j] = strtod(fpnumber, NULL) / (double)numberFFTs[i];
                    }
                    fgets(buf, 120, infile);  // CR
                }
                hdc = GetDC(mainHWND);
                drawGraphs(FFTsizes, FFTsizes, hdc);
                ReleaseDC(mainHWND, hdc); 
                sprintf(fileMessage, "Search for another graph?");
                sprintf(buf3, " Loaded graph %d of %d from file %s", count2, grafCount, fileToRead);
                if (count2 < grafCount)
                {
                    mbYNC = MessageBox(mainHWND, fileMessage, buf3,
                                        MB_ICONQUESTION | MB_YESNO);
                    if (mbYNC == IDNO)
                    {
                       fin = TRUE;
                       gotGraf = 1;
                    }
                    else
                    {
                        gotGraf = 2;
                    }
                }
                else
                {
                    fin = TRUE;
                    gotGraf = 1;
                }
            }
        }
    }
    sprintf(buf3, " Loading graph from file %s", fileToRead);
    if (gotGraf == 0)
    {
        MessageBox(NULL, "No graph found", buf3, MB_ICONWARNING | MB_OK);
    }
    else if (gotGraf == 2)
    {
        MessageBox(NULL, "No more graphs found", buf3, MB_ICONWARNING | MB_OK);
    }    
    fclose(infile);
    return TRUE;
     
} // loadFile()


// bit reverse table

void bitrev (int t)
{
  int index;
  int point;
  int m;
  int p;
  int Nsize, Tsize;

  if(t > TsizeL2)
         Tsize = TsizeL2;
  else
          Tsize = t;

  Nsize = (int)pow(2,(double)Tsize);

  for (index=0; index<Nsize; index++)
  {
    p = 0;
    m = index;
    for (point=0; point<Tsize; point++)
    {
      p = p + p;
      p = p + (m & 1);
      m = m >> 1;
    }
    table[index] = p;
  }

} // bit reverse table


// create exponent table

void createExpTableSP (int n, int sign)
{
    int   index, Nsize;
    float const0;
    float rindex;
    float fsign;
    float pi2;
    
    fsign = (float)sign;
    pi2 = (float)2.0 * PISP;  
    const0 = (pi2 * fsign)  / (float)n;
    for (index=0; index<n/2; index++)
    {
        rindex = (float)index;
        cExpSP[index*2]   = (float)cos (const0 * rindex);
        cExpSP[index*2+1] = (float)sin (const0 * rindex);
    }
    Nsize = (int)pow(2,(float)TsizeL2);
    if(n < Nsize) Nsize = n;
    const0 = (pi2 * fsign)  / (float)Nsize;
    for (index=0; index<Nsize/2; index++)
    {
        rindex = (float)index;
        dExpSP[index*2]      = (float)cos (const0 * rindex);
        dExpSP[(index*2)+1]  = (float)sin (const0 * rindex);
    }
}
void createExpTableDP (int n, int sign)
{
    int   index, Nsize;
    double const0;
    double rindex;
    double fsign;
    double pi2;
    
    fsign = (double)sign;
    pi2 = 2.0 * PIDP;  
    const0 = (pi2 * fsign)  / (double)n;
    for (index=0; index<n/2; index++)
    {
        rindex = (double)index;
        cExpDP[index*2]   = cos (const0 * rindex);
        cExpDP[index*2+1] = sin (const0 * rindex);
    }
    Nsize = (int)pow(2,(double)TsizeL2);
    if(n < Nsize) Nsize = n;
    const0 = (pi2 * fsign)  / (double)Nsize;
    for (index=0; index<Nsize/2; index++)
    {
        rindex = (double)index;
        dExpDP[index*2]     = cos (const0 * rindex);
        dExpDP[(index*2)+1] = sin (const0 * rindex);
    }
} // create exponent table

// create data to avoid errors

void createComplexDataSP (int n, float synthFreqI, float synthFreqQ)
{
  float  constI, constQ;
  float  rindex;
  int   index;
 
  constI = ((float)2.0  * PISP * synthFreqI) / (float)n;
  constQ = ((float)2.0  * PISP * synthFreqQ) / (float)n;

  for (index=0; index<n; index++)
  {
    rindex = (float)index;
    cSourceSP[index*2]     = (float)cos (constI * rindex);
    cSourceSP[(index*2)+1] = (float)sin (constQ * rindex);
  }
}

void createComplexDataDP (int n, double synthFreqI, double synthFreqQ)
{
  double  constI, constQ;
  double  rindex;
  int   index;
 
  constI = (2.0  * PIDP * synthFreqI) / (double)n;
  constQ = (2.0  * PIDP * synthFreqQ) / (double)n;

  for (index=0; index<n; index++)
  {
    rindex = (double)index;
    cSourceDP[index*2]     = cos (constI * rindex);
    cSourceDP[(index*2)+1] = sin (constQ * rindex);
  }
}

void fftSP2 (int t, int n)
{
  float  inverse_n;
  int   d, d2;
  int   index;
  int   point;
  int   j; 
  int   jinc;
  float  temp1, temp2;

  inverse_n = (float)1.0 / (float)n;
  
    for (index=0; index<n; index++)
    {
            point = table[index];
            if (point >= index)
            {
              temp1  = cSourceSP[index*2]     * inverse_n;
              temp2  = cSourceSP[(index*2)+1] * inverse_n;
              cSourceSP[index*2]     = cSourceSP[point*2]     * inverse_n;
              cSourceSP[(index*2)+1] = cSourceSP[(point*2)+1] * inverse_n;
              cSourceSP[point*2]     = temp1;
              cSourceSP[(point*2)+1] = temp2;
            }
            
    }

    d = 1;
    for(index=1;index<=t;index++)
    {
        d2 = d + d;
        jinc = 2;
        if (d > 2) jinc = 4;
        for (j=0; j<d; j=j+jinc)
        {

            ExpReal1SP = cExpSP[(n / d2) * j * 2];
            ExpImag1SP = cExpSP[(n / d2) * j * 2+1];
            ExpReal2SP = cExpSP[(n / d2) * (j+1)*2];
            ExpImag2SP = cExpSP[(n / d2) * (j+1)*2+1];

            if (d > 2)
            { 
                ExpReal3SP = cExpSP[(n / d2) * (j+2)*2];
                ExpImag3SP = cExpSP[(n / d2) * (j+2)*2+1];    
                ExpReal4SP = cExpSP[(n / d2) * (j+3)*2];
                ExpImag4SP = cExpSP[(n / d2) * (j+3)*2+1];
            }

            if (d == 1)
            {
                if (has3DNow)
                {
                    ass1SP3DNow(n, j, d2, (int)cSourceSP);
                }
                else
                {
                    ass1SP(n, j, d2, (int)cSourceSP);
                }
            }
            else if (d == 2)
            {
                if (doSSE1)
                {
                    ass2SPSSE(n, j, d2, (int)cSourceSP);
                }
                else if (has3DNow)
                {
                    ass2SP3DNow(n, j, d2, (int)cSourceSP);
                }
                else
                {
                    ass2SP(n, j, d2, (int)cSourceSP);
                }
            }
            else
            {
                if (doSSE1)
                {
                    ass4SPSSE(n, j, d2, (int)cSourceSP);
                }
                else if (has3DNow)
                {
                    ass4SP3DNow(n, j, d2, (int)cSourceSP);
                }
                else
                {
                    ass4SP(n, j, d2, (int)cSourceSP);
                }
            }
        }
        d = d2;
    }
} // fftSP2


void fftSP (int t, int n)
{
    float  inverse_n;
    float  *pData;
    int   index;
    int   d, d2;
    int   point, Nsize, Tsize, seg, NumSeg, offset, RevSeg;
    int   j, k, jinc;

    inverse_n = (float)1.0 / (float)n;

    if(t >= TsizeL2)
       Tsize = TsizeL2;
    else
        Tsize = t;

    Nsize  = (int)pow(2,(float)Tsize);
    NumSeg = n / Nsize;

    //************************************************************
    // Break the work up into NumSeg work units of size TsizeL2 
    //************************************************************

    for(seg=0;seg<NumSeg;seg++)
    {
        RevSeg = 0;
        for(k=0;k<(t-Tsize);k++)
                RevSeg = (RevSeg << 1) | ((seg >> k) & 0x1);

        pData = cDataSP+(RevSeg*Nsize*2);
        offset = seg * 2;

        // bit reverse each set
        for (index=0; index<Nsize; index++)
        {
            point = table[index];
            if (point >= index)
            {
                pData[index*2]     = cSourceSP[(point*2*NumSeg)+offset]   * inverse_n;
                pData[(index*2)+1] = cSourceSP[(point*2*NumSeg)+offset+1] * inverse_n;       
                pData[point*2]     = cSourceSP[(index*2*NumSeg)+offset]   * inverse_n;
                pData[(point*2)+1] = cSourceSP[(index*2*NumSeg)+offset+1] * inverse_n;
            }
        }

        // transform each set
        d = 1;
        for(index=1;index<=Tsize;index++)
        {
            d2 = d + d;
            jinc = 2;
            if (d > 2) jinc = 4;
            for (j=0; j<d; j=j+jinc)
            {

                ExpReal1SP = dExpSP[(Nsize / d2) * j * 2];
                ExpImag1SP = dExpSP[(Nsize / d2) * j * 2+1];
                ExpReal2SP = dExpSP[(Nsize / d2) * (j+1)*2];
                ExpImag2SP = dExpSP[(Nsize / d2) * (j+1)*2+1];

                if (d > 2)
                { 
                    ExpReal3SP = dExpSP[(Nsize / d2) * (j+2)*2];
                    ExpImag3SP = dExpSP[(Nsize / d2) * (j+2)*2+1];    
                    ExpReal4SP = dExpSP[(Nsize / d2) * (j+3)*2];
                    ExpImag4SP = dExpSP[(Nsize / d2) * (j+3)*2+1];
                }

                if (d == 1)
                {
                    if (has3DNow)
                    {
                        ass1SP3DNow(Nsize, j, d2, (int)pData);
                    }
                    else
                    {
                        ass1SP(Nsize, j, d2, (int)pData);
                    }
                }
                else if (d == 2)
                {
                    if (doSSE1)
                    {
                        ass2SPSSE(Nsize, j, d2, (int)pData);
                    }
                    else if (has3DNow)
                    {
                        ass2SP3DNow(Nsize, j, d2, (int)pData);
                    }
                    else
                    {
                        ass2SP(Nsize, j, d2, (int)pData);
                    }
                }
                else
                {
                    if (doSSE1)
                    {
                        ass4SPSSE(Nsize, j, d2, (int)pData);
                    }
                    else if (has3DNow)
                    {
                        ass4SP3DNow(Nsize, j, d2, (int)pData);
                    }
                    else
                    {
                        ass4SP(Nsize, j, d2, (int)pData);
                    }
                }                            
            }
            d = d2;
        }
    }

    //***********************************************************
    // This will recombine the above work units (if necessary) 
    //  Only run when t > TsizeL2 eg TsizeL2 = 13 run at size 16384
    //**********************************************************

   for(index=TsizeL2+1; index<=t; index++)
   {
      d2 = d + d;
      for (j=0; j<d; j=j+4)
      {
        ExpReal1SP = cExpSP[(n / d2 * j)*2];
        ExpImag1SP = cExpSP[(n / d2 * j)*2+1];
        ExpReal2SP = cExpSP[(n / d2 * (j+1))*2];
        ExpImag2SP = cExpSP[(n / d2 * (j+1))*2+1];
        ExpReal3SP = cExpSP[(n / d2 * (j+2))*2];
        ExpImag3SP = cExpSP[(n / d2 * (j+2))*2+1];
        ExpReal4SP = cExpSP[(n / d2 * (j+3))*2];
        ExpImag4SP = cExpSP[(n / d2 * (j+3))*2+1];

        if (doSSE1)
        {
            ass4SPSSE(n, j, d2, (int)cDataSP);
        }
        else if (has3DNow)
        {
            ass4SP3DNow(n, j, d2, (int)cDataSP);
        }
        else
        {
            ass4SP(n, j, d2, (int)cDataSP);
        }
      }
      d = d2;
   }

} // fftSP

void fftDP2 (int t, int n)
{
  double  inverse_n;
  int   d, d2;
  int   index;
  int   point;
  int   j; 
  int   jinc;
  double  temp1, temp2;

  inverse_n = 1.0 / (double)n;
  cExpDPA[16] = inverse_n;
  cExpDPA[17] = inverse_n;

  if (doSSE2)
  {
    for (index=0; index<n; index++)
    {
            point = table[index];
            if (point >= index)
            {
              vDP[1]         = _mm_mul_pd(cDP[index] , vDP[8]);
              cDP[index]     = _mm_mul_pd(cDP[point] , vDP[8]);
              cDP[point]     = vDP[1]; 
            }
    }
  }
  else
  {
    for (index=0; index<n; index++)
    {
            point = table[index];
            if (point >= index)
            {
              temp1  = cSourceDP[index*2] * inverse_n;
              temp2  = cSourceDP[(index*2)+1] * inverse_n;
              cSourceDP[index*2] = cSourceDP[point*2] * inverse_n;
              cSourceDP[(index*2)+1] = cSourceDP[(point*2)+1] * inverse_n;
              cSourceDP[point*2]     = temp1;
              cSourceDP[(point*2)+1] = temp2;
            }
    }
  }
 
    d = 1;
    for(index=1;index<=t;index++)
    {
        d2 = d + d;
        jinc = 2;
        if (d > 2) jinc = 4;
        for (j=0; j<d; j=j+jinc)
        {

            ExpReal1DP = cExpDP[(n / d2) * j * 2];
            ExpImag1DP = cExpDP[(n / d2) * j * 2+1];
            ExpReal2DP = cExpDP[(n / d2) * (j+1)*2];
            ExpImag2DP = cExpDP[(n / d2) * (j+1)*2+1];

            if (d > 2)
            { 
                ExpReal3DP = cExpDP[(n / d2) * (j+2)*2];
                ExpImag3DP = cExpDP[(n / d2) * (j+2)*2+1];    
                ExpReal4DP = cExpDP[(n / d2) * (j+3)*2];
                ExpImag4DP = cExpDP[(n / d2) * (j+3)*2+1];
            }

            if (d == 1)
            {
                if (doSSE2)
                {
                    ass1DPSSE2(n, j, d2, (int)cSourceDP);
                }
                else
                {
                    ass1DP(n, j, d2, (int)cSourceDP);
                }
            }
            else if (d == 2)
            {
                if (doSSE2)
                {
                    ass2DPSSE2(n, j, d2, (int)cSourceDP);
                }
                else
                {
                    ass2DP(n, j, d2, (int)cSourceDP);
                }
            }
            else
            {
                if (doSSE2)
                {
                    ass4DPSSE2(n, j, d2, (int)cSourceDP);
                }
                else
                {
                    ass4DP(n, j, d2, (int)cSourceDP);
                }
            }
        }
        d = d2;
    }
} // fftDP2

void fftDP (int t, int n)
{
    double  inverse_n;
    double  *pData;
    __m128d  *pDP;
    
    int   d, d2;
    int   index;
    int   point, Nsize, Tsize, seg, NumSeg, offset, RevSeg;
    int   offo2;
    int   j, k, jinc;

    inverse_n = 1.0 / (double)n;

    cExpDPA[16] = inverse_n;
    cExpDPA[17] = inverse_n;

    if(t >= TsizeL2)
    {
       Tsize = TsizeL2;
    }
    else
    {
        Tsize = t;
    }
    Nsize  = (int)pow(2,(double)Tsize);
    NumSeg = n / Nsize;
    
    //************************************************************
    // Break the work up into NumSeg work units of size TsizeL2 
    //************************************************************

    for(seg=0;seg<NumSeg;seg++)
    {
        RevSeg = 0;
        for(k=0;k<(t-Tsize);k++)
                RevSeg = (RevSeg << 1) | ((seg >> k) & 0x1);

        pData = cDataDP+(RevSeg*Nsize*2);
        offset = seg * 2;
        offo2  = seg;
        pDP = (__m128d *)pData;

        // bit reverse each set
/*
This is a little slower

        if (doSSE2)
        {
            for (index=0; index<Nsize; index++)
            {
                point = table[index];
                if (point >= index)
                {
                    pDP[index]   = _mm_mul_pd(cDP[point*NumSeg+offo2] , vDP[8]);
                    pDP[point]   = _mm_mul_pd(cDP[index*NumSeg+offo2] , vDP[8]);
                }
            }
        }
        else
*/
        {
            for (index=0; index<Nsize; index++)
            {
                point = table[index];
                if (point >= index)
                {
                    pData[index*2]     = cSourceDP[(point*2*NumSeg)+offset]   * inverse_n;
                    pData[(index*2)+1] = cSourceDP[(point*2*NumSeg)+offset+1] * inverse_n;       
                    pData[point*2]     = cSourceDP[(index*2*NumSeg)+offset]   * inverse_n;
                    pData[(point*2)+1] = cSourceDP[(index*2*NumSeg)+offset+1] * inverse_n;
                }
            }
        }
        
        // transform each set
        d = 1;
        for(index=1;index<=Tsize;index++)
        {
            d2 = d + d;
            jinc = 2;
            if (d > 2) jinc = 4;
            for (j=0; j<d; j=j+jinc)
            {

                ExpReal1DP = dExpDP[(Nsize / d2) * j * 2];
                ExpImag1DP = dExpDP[(Nsize / d2) * j * 2+1];
                ExpReal2DP = dExpDP[(Nsize / d2) * (j+1)*2];
                ExpImag2DP = dExpDP[(Nsize / d2) * (j+1)*2+1];

                if (d > 2)
                { 
                    ExpReal3DP = dExpDP[(Nsize / d2) * (j+2)*2];
                    ExpImag3DP = dExpDP[(Nsize / d2) * (j+2)*2+1];    
                    ExpReal4DP = dExpDP[(Nsize / d2) * (j+3)*2];
                    ExpImag4DP = dExpDP[(Nsize / d2) * (j+3)*2+1];
                }

                if (d == 1)
                {
                    if (doSSE2)
                    {
                        ass1DPSSE2(Nsize, j, d2, (int)pData);
                    }
                    else
                    {
                        ass1DP(Nsize, j, d2, (int)pData);
                    }
                }
                else if (d == 2)
                {
                    if (doSSE2)
                    {
                        ass2DPSSE2(Nsize, j, d2, (int)pData);
                    }
                    else
                    {
                        ass2DP(Nsize, j, d2, (int)pData);
                    }
                }
                else
                {
                    if (doSSE2)
                    {
                        ass4DPSSE2(Nsize, j, d2, (int)pData);
                    }
                    else
                    {
                        ass4DP(Nsize, j, d2, (int)pData);
                    }
                }                            
            }
            d = d2;
         }
    }

    //***********************************************************
    // This will recombine the above work units (if necessary) 
    //  Only run when t > TsizeL2 eg TsizeL2 = 13 run at size 16384
    //**********************************************************

   for(index=TsizeL2+1; index<=t; index++)
   {
      d2 = d + d;
      for (j=0; j<d; j=j+4)
      {
        ExpReal1DP = cExpDP[(n / d2 * j)*2];
        ExpImag1DP = cExpDP[(n / d2 * j)*2+1];
        ExpReal2DP = cExpDP[(n / d2 * (j+1))*2];
        ExpImag2DP = cExpDP[(n / d2 * (j+1))*2+1];
        ExpReal3DP = cExpDP[(n / d2 * (j+2))*2];
        ExpImag3DP = cExpDP[(n / d2 * (j+2))*2+1];
        ExpReal4DP = cExpDP[(n / d2 * (j+3))*2];
        ExpImag4DP = cExpDP[(n / d2 * (j+3))*2+1];
        if (doSSE2)
        {
            ass4DPSSE2(n, j, d2, (int)cDataDP);
        }
        else
        {
            ass4DP(n, j, d2, (int)cDataDP);
        }
      }
      d = d2;
   }

} // fftDP


void SPCalculate()
{
    HDC   hdc;
    int i, j, t, n, sign;
    float synthFreq;
    int  Tmin, Tmax;
    
    keyPress = GetAsyncKeyState(VK_ESCAPE);

    testRunning = TRUE;

    if (tpass == 0 && jpass == 0)
    {
        for (i=0; i<FFTsizes; i++)
        {
            for (j=0; j<maxruns; j++)
            {
                SPFFT[i][j] = 0;
                DPFFT[i][j] = 0;
            }
        }
    }
    Tmin = minFFTRun + 10;
    Tmax = maxFFTRun + 10;
    

    t = tpass + Tmin;
    {
        runningTest = t - 10;
        n = (int)pow(2,(double)t);
        nSP = n;    
        bitrev (t); 
        sign = -1; 
        createExpTableSP(n, sign);
        synthFreq = (float)n/16;
        j = jpass;
        {
            if (flash1 == 1)
            {
                flash1 = 2;
            }
            else
            {
                flash1 = 1;
            }
            createComplexDataSP(n, synthFreq, synthFreq);

            getStartTime();
            if (t < TsizeLL)
            {    
               fftSP2 (t, n);
            }
            else
            {
               fftSP (t, n);
            } 
            getRunSecs();
            SPFFT[runningTest][j] = runSecs * 1000 / (double)(n / 1024);
            hdc = GetDC(mainHWND);
            drawGraphs(FFTsizes, FFTsizes, hdc);
            ReleaseDC(mainHWND, hdc); 
            if (GetAsyncKeyState(VK_ESCAPE) != keyPress)
            {
                escPressed = TRUE;
            }
        }
        if (!escPressed &&  runSecs * 2 > (double)maxSecs
                                && t != Tmax && j == FFTpasses - 1)
        {
            sprintf (amsg, "Next tests likely to take >%d seconds\n"
                           "Continue or Cancel next tests?", maxSecs); 
            if (MessageBox(mainHWND, amsg, "Test Time Too Long?",
                    MB_ICONQUESTION | MB_OKCANCEL) == IDCANCEL)
             {
                 tpass = maxFFTRun + 1;
             }
        }
    }    
} // SPCalculate()

void DPCalculate()
{
    HDC   hdc;
    int i, j, t, n, sign;
    float synthFreq;
    int  Tmin, Tmax;
    
    keyPress = GetAsyncKeyState(VK_ESCAPE);

    testRunning = TRUE;
    testRunning = TRUE;

    if (tpass == 0 && jpass == 0)
    {
        for (i=0; i<FFTsizes; i++)
        {
            for (j=0; j<FFTpasses; j++)
            {
                DPFFT[i][j] = 0;
            }
        }
    }
    
    Tmin = minFFTRun + 10;
    Tmax = maxFFTRun + 10;
    
    t = tpass + Tmin;
    {
        runningTest = t - 10;
        n = (int)pow(2,(double)t);
        nDP = n;    
        bitrev (t); 
        sign = -1; 
        createExpTableDP(n, sign);
        synthFreq = (float)n/16;
        j = jpass;
        {
            if (flash2 == 1)
            {
                flash2 = 2;
            }
            else
            {
                flash2 = 1;
            }
            createComplexDataDP(n, synthFreq, synthFreq);
            getStartTime();
            if (t < TsizeLL)
            {    
            fftDP2 (t, n); 
            }
            else
            {
            fftDP (t, n); 
            }            
            getRunSecs();
            DPFFT[runningTest][j] = runSecs * 1000 / (double)(n / 1024);
            hdc = GetDC(mainHWND);
            drawGraphs(FFTsizes, FFTsizes, hdc);
            ReleaseDC(mainHWND, hdc); 
            if (GetAsyncKeyState(VK_ESCAPE) != keyPress)
            {
                escPressed = TRUE;
            }
        }
        if (!escPressed &&  runSecs * 2 > (double)maxSecs
                          && t != Tmax && j == FFTpasses - 1)
        {
            sprintf (amsg, "Next tests likely to take >%d seconds\n"
                           "Continue or Cancel next tests?", maxSecs); 
            if (MessageBox(mainHWND, amsg, "Test Time Too Long?",
                    MB_ICONQUESTION | MB_OKCANCEL) == IDCANCEL)
             {
                 tpass = maxFFTRun + 1;
             }
        }
    }    
} // DPCalculate()


void Compare0()
{
    int i, j;
    
    for (i=0; i<FFTsizes; i++)
    {
        for (j=0; j<maxruns; j++)
        {
            SP1[i][j] = 0;
            SP2[i][j] = 0;
            DP1[i][j] = 0;
            DP2[i][j] = 0;
        }
    }
}

void Compare1()
{
    int i, j;

    for (i=0; i<FFTsizes; i++)
    {
        for (j=0; j<maxruns; j++)
        {
            SP1[i][j] = SPFFT[i][j];
            DP1[i][j] = DPFFT[i][j];
        }
        miniSPA[i] = miniSP[i];
        miniDPA[i] = miniDP[i];
    }
    if (strlen(configData1) < 49)
    {
        strcpy(config1, configData1);
    }
    else
    {
        strncpy(config1, configData1, 48);
    } 
    strcpy(config3, configData2);
    vers1 = vers2; 
}

void Compare2()
{
    int i, j;
    HDC   hdc;
    
    for (i=0; i<FFTsizes; i++)
    {
        for (j=0; j<maxruns; j++)
        {
            SP2[i][j] = SPFFT[i][j];
            DP2[i][j] = DPFFT[i][j];
            if (SP1[i][j] * SP2[i][j] > 0)
            {
                SPFFT[i][j] = SP2[i][j] / SP1[i][j];
            }
            else
            {
                SPFFT[i][j] = 0;
            }
            if (DP1[i][j] * DP2[i][j] > 0)
            {
                DPFFT[i][j] = DP2[i][j] / DP1[i][j];
            }
            else
            {
                DPFFT[i][j] = 0;
            }
        }
        miniSPB[i] = miniSP[i];
        miniDPB[i] = miniDP[i];
    }
    if (strlen(configData1) < 49)
    {
        strcpy(config2, configData1);
    }
    else
    {
        strncpy(config2, configData1, 48);
    } 
    strcpy(configData1, config3); 
    hdc = GetDC(mainHWND);
    compareTwo = TRUE;
    drawGraphs(FFTsizes, FFTsizes, hdc);
    ReleaseDC(mainHWND, hdc); 
}

BOOL compareTxt()
{
     int i, j, n;
     double ratio;
     double minSP1[11];
     double minDP1[11];
     double maxy;
     double maxz;
     double miny;
     int  pos1;
     int  pos2;
     int  pos3;
     int  pos4;
     int  pos5;
     FILE *outfile;

     SetCurrentDirectory(currentDir);
     outfile = fopen("Compare2.txt","a+");
     if (outfile == NULL)
     {
         MessageBox(NULL, "Cannot open compare file", "Compare2.txt",
                                 MB_ICONWARNING | MB_OK);
         return FALSE;
     }
     fprintf (outfile, "\n\n ###########################################################\n");
     fprintf (outfile, "                 FFTGraf Test Speed Comparison\n\n");
     fprintf (outfile, " A %s\n", config1);
     fprintf (outfile, " B %s\n", config2);
     fprintf (outfile, " A %s\n", configData1);
     fprintf (outfile, " B %s\n\n", configData2);
     if (vers1 ==  vers2)
     {
         if (vers1 == 1)
         {
             fprintf (outfile, " Comparison of Original Version %d Results\n\n", vers1);
         }
         else
         {
             fprintf (outfile, " Comparison of RAM Burst Speed Tuned Version %d Results\n\n", vers1);
         }
     }
     else
     {
         fprintf (outfile, " Comparison WARNING A = Version %d B = Version %d\n\n ", vers1, vers2);
     }

     fprintf (outfile, " Minimum Millisecs Ratio B/A. If > 1.0 A is faster\n\n");

     maxy = 0;
     miny = 1000000;
     for (i=0; i<FFTsizes; i++)
     {
        if (miniSPA[i] > 0)
        {
            minSP1[i] = miniSPB[i] / miniSPA[i];
            if (minSP1[i] > maxy) maxy = minSP1[i]; 
            if (minSP1[i] > 0 && minSP1[i] < miny) miny = minSP1[i]; 
        }
        else
        {
            minSP1[i] = 0;
        }
        if (miniDPA[i] > 0)
        {
            minDP1[i] = miniDPB[i] / miniDPA[i];
            if (minDP1[i] > maxy) maxy = minDP1[i]; 
            if (minSP1[i] > 0 && minDP1[i] < miny) miny = minDP1[i]; 
        }
        else
        {
            minDP1[i] = 0;
        }
     }
     maxz = maxy;

     fprintf (outfile, " Minimum Ratio =  %6.4f. Maximum Ratio = %6.4f\n\n", miny, maxy);
          
     if (maxy > 0) maxy = 60 / maxy;
     pos1 = (int)(maxy + 0.5); 
     pos2 = (int)(2 * maxy + 0.5);
     pos3 = (int)(3 * maxy + 0.5);
     pos4 = (int)(4 * maxy + 0.5);
     pos5 = (int)(5 * maxy + 0.5);

     fprintf (outfile, " Single Precision FFTs\n");
     fprintf (outfile, " Memory 16KB, 32KB, 64KB, 128KB, 256KB+, 512KB+, 1MB+ ... 16MB+\n\n");

     for (i=0; i<FFTsizes; i++)
     {
        ratio = minSP1[i];
        n = (int)(ratio * maxy + 0.5);
        if (ratio > 9.999999)
        {
            fprintf (outfile, " >=10  ");
        }
        else if (ratio == 0)
        {
            fprintf (outfile, " N/A  ");
        }
        else
        {
            fprintf (outfile, " %4.2f  ", ratio);
        }
        for (j=1; j<n+1; j++)
        {
            if (maxz > 9.999999)
            {
                fprintf (outfile, "#");
            }
            else if (j == pos1)
            {
                fprintf (outfile, "1");
            }
            else if (j == pos2)
            {
                fprintf (outfile, "2");
            }
            else if (j == pos3)
            {
                fprintf (outfile, "3");
            }
            else if (j == pos4)
            {
                fprintf (outfile, "4");
            }
            else if (j == pos5)
            {
                fprintf (outfile, "5");
            }
            else
            {
                fprintf (outfile, "#");
            }
        }
        fprintf (outfile, "\n");
     }
     fprintf (outfile, "\n");
 
     fprintf (outfile, "\n Double Precision FFTs\n");
     fprintf (outfile, " Memory 28KB, 56KB, 112KB, 224KB, 448KB+, 896KB+ ... 28MB+\n\n");
     
     for (i=0; i<FFTsizes; i++)
     {
        ratio = minDP1[i];
        n = (int)(ratio * maxy + 0.5);
        if (ratio > 9.999999)
        {
            fprintf (outfile, " >=10  ");
        }
        else if (ratio == 0)
        {
            fprintf (outfile, " N/A  ");
        }
        else
        {
            fprintf (outfile, " %4.2f  ", ratio);
        }
        for (j=1; j<n+1; j++)
        {
            if (maxz > 9.999999)
            {
                fprintf (outfile, "#");
            }
            else if (j == pos1)
            {
                fprintf (outfile, "1");
            }
            else if (j == pos2)
            {
                fprintf (outfile, "2");
            }
            else if (j == pos3)
            {
                fprintf (outfile, "3");
            }
            else if (j == pos4)
            {
                fprintf (outfile, "4");
            }
            else if (j == pos5)
            {
                fprintf (outfile, "5");
            }
            else
            {
                fprintf (outfile, "#");
            }
        }
        fprintf (outfile, "\n");
     }
     fprintf (outfile, "\n");
     fflush (outfile);
     fclose(outfile);
     return TRUE;

} // compareTxt

void checkNoiseSP()
{
    int k;
    float sqOne;
    float MaxNoise = 0.0;
    float AveNoise = 0.0;
    
    for(k=1;k<nSP;k++)
    {
        if(k != nSP/16)
        {
            sqOne = cDataSP[k*2] * cDataSP[k*2] + cDataSP[(k*2)+1] * cDataSP[(k*2)+1];
            if(sqOne > MaxNoise)
                    MaxNoise = sqOne;
            AveNoise = AveNoise + sqOne;
        }
    }
    AveNoise = AveNoise / (nSP-2);
    sqOne = cDataSP[(nSP/16)*2] * cDataSP[(nSP/16)*2] + cDataSP[((nSP/16)*2)+1] * cDataSP[((nSP/16)*2)+1];
    sprintf(checkSP, " Checks SP %14e  %10e  %10e", sqOne, MaxNoise, AveNoise);
}

void checkNoiseDP()
{
    int k;
    double sqOne;
    double MaxNoise = 0.0;
    double AveNoise = 0.0;
    
    for(k=1;k<nDP;k++)
    {
        if(k != nDP/16)
        {
            sqOne = cDataDP[k*2] * cDataDP[k*2] + cDataDP[(k*2)+1] * cDataDP[(k*2)+1];
            if(sqOne > MaxNoise)
                    MaxNoise = sqOne;
            AveNoise = AveNoise + sqOne;
        }
    }
    AveNoise = AveNoise / (nDP-2);
    sqOne = cDataDP[(nDP/16)*2] * cDataDP[(nDP/16)*2] + cDataDP[((nDP/16)*2)+1] * cDataDP[((nDP/16)*2)+1];
    sprintf(checkDP, " Checks DP %14e  %10e  %10e", sqOne, MaxNoise, AveNoise);
}

void ass1DP(int n, int j, int d2,  int cDataDP)
{
    n = n * 8;
    j = j * 8;
    d2 = d2 * 8;
    __asm
    {
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataDP            ; not lea
        add     esi, edx                ; address cDataDP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
    lp: mov     eax, esi                ; address cDataDP[j2]
        add     eax, ecx                ; address cDataDP[jk or j2 + k]

        fld     ExpImag1DP
        fld     qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+1] * ExpImag
        fld     ExpReal1DP
        fld     qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1] * ExpReal
        fld     qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax]         ; cDataDP[j2+k]
        fsub    st,st(1)                ; cDataDP[j2+k] - tempReal
        fstp    qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]
        fld     qword ptr [eax+8]       ; cDataDP[j2+k+1]
        fsub    st,st(2)                ; cDataDP[j2+k+1] - tempImag
        fstp    qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1]
        fadd    qword ptr [eax]         ; cDataDP[j2+k] + tempReal
        fstp    qword ptr [eax]         ; cDataDP[j2+k]
        fadd    qword ptr [eax+8]       ; cDataDP[j2+k+1] + tempImag
        fstp    qword ptr [eax+8]       ; cDataDP[j2+k+1]

        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} 
void ass2DP(int n, int j, int d2,  int cDataDP)
{
    n = n * 8;
    j = j * 8;
    d2 = d2 * 8;
    __asm
    {
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataDP            ; not lea
        add     esi, edx                ; address cDataDP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
    lp: mov     eax, esi                ; address cDataDP[j2]
        add     eax, ecx                ; address cDataDP[jk or j2 + k]

        fld     ExpImag1DP                ; PART 1
        fld     qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+1] * ExpImag
        fld     ExpReal1DP
        fld     qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1] * ExpReal
        fld     qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax]         ; cDataDP[j2+k]
        fsub    st,st(1)                ; cDataDP[j2+k] - tempReal
        fstp    qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]
        fld     qword ptr [eax+8]       ; cDataDP[j2+k+1]
        fsub    st,st(2)                ; cDataDP[j2+k+1] - tempImag
        fstp    qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1]
        fadd    qword ptr [eax]         ; cDataDP[j2+k] + tempReal
        fstp    qword ptr [eax]         ; cDataDP[j2+k]
        fadd    qword ptr [eax+8]       ; cDataDP[j2+k+1] + tempImag
        fstp    qword ptr [eax+8]       ; cDataDP[j2+k+1]

        fld     ExpImag2DP              ; PART 2
        fld     qword ptr [eax+ebx+24]  ; cDataDP[j2+k+d2+3]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+3] * ExpImag
        fld     ExpReal2DP
        fld     qword ptr [eax+ebx+16]  ; cDataDP[j2+k+d2+2]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2+2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+24]  ; cDataDP[j2+k+d2+3] * ExpReal
        fld     qword ptr [eax+ebx+16]  ; cDataDP[j2+k+d2+2]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2+2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax+16]      ; cDataDP[j2+k+2]
        fsub    st,st(1)                ; cDataDP[j2+k+2] - tempReal
        fstp    qword ptr [eax+ebx+16]  ; cDataDP[j2+k+d2+2]
        fld     qword ptr [eax+24]      ; cDataDP[j2+k+3]
        fsub    st,st(2)                ; cDataDP[j2+k+3] - tempImag
        fstp    qword ptr [eax+ebx+24]  ; cDataDP[j2+k+d2+3]
        fadd    qword ptr [eax+16]      ; cDataDP[j2+k+2] + tempReal
        fstp    qword ptr [eax+16]      ; cDataDP[j2+k+2]
        fadd    qword ptr [eax+24]      ; cDataDP[j2+k+3] + tempImag
        fstp    qword ptr [eax+24]      ; cDataDP[j2+k+3]

        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} 

void ass4DP(int n, int j, int d2,  int cDataDP)
{
    n = n * 8;
    j = j * 8;
    d2 = d2 * 8;
    __asm
    {
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataDP              ; not lea
        add     esi, edx                ; address cDataDP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
    lp: mov     eax, esi                ; address cDataDP[j2]
        add     eax, ecx                ; address cDataDP[jk or j2 + k]

        fld     ExpImag1DP              ; PART 1
        fld     qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+1] * ExpImag
        fld     ExpReal1DP
        fld     qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1] * ExpReal
        fld     qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax]         ; cDataDP[j2+k]
        fsub    st,st(1)                ; cDataDP[j2+k] - tempReal
        fstp    qword ptr [eax+ebx]     ; cDataDP[j2+k+d2]
        fld     qword ptr [eax+8]       ; cDataDP[j2+k+1]
        fsub    st,st(2)                ; cDataDP[j2+k+1] - tempImag
        fstp    qword ptr [eax+ebx+8]   ; cDataDP[j2+k+d2+1]
        fadd    qword ptr [eax]         ; cDataDP[j2+k] + tempReal
        fstp    qword ptr [eax]         ; cDataDP[j2+k]
        fadd    qword ptr [eax+8]       ; cDataDP[j2+k+1] + tempImag
        fstp    qword ptr [eax+8]       ; cDataDP[j2+k+1]

        fld     ExpImag2DP              ; PART 2
        fld     qword ptr [eax+ebx+24]  ; cDataDP[j2+k+d2+3]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+3] * ExpImag
        fld     ExpReal2DP
        fld     qword ptr [eax+ebx+16]  ; cDataDP[j2+k+d2+2]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2+2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+24]  ; cDataDP[j2+k+d2+3] * ExpReal
        fld     qword ptr [eax+ebx+16]  ; cDataDP[j2+k+d2+2]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2+2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax+16]      ; cDataDP[j2+k+2]
        fsub    st,st(1)                ; cDataDP[j2+k+2] - tempReal
        fstp    qword ptr [eax+ebx+16]  ; cDataDP[j2+k+d2+2]
        fld     qword ptr [eax+24]      ; cDataDP[j2+k+3]
        fsub    st,st(2)                ; cDataDP[j2+k+3] - tempImag
        fstp    qword ptr [eax+ebx+24]  ; cDataDP[j2+k+d2+3]
        fadd    qword ptr [eax+16]      ; cDataDP[j2+k+2] + tempReal
        fstp    qword ptr [eax+16]      ; cDataDP[j2+k+2]
        fadd    qword ptr [eax+24]      ; cDataDP[j2+k+3] + tempImag
        fstp    qword ptr [eax+24]      ; cDataDP[j2+k+3]

        fld     ExpImag3DP              ; PART 3
        fld     qword ptr [eax+ebx+40]  ; cDataDP[j2+k+d2+5]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+5] * ExpImag
        fld     ExpReal3DP
        fld     qword ptr [eax+ebx+32]  ; cDataDP[j2+k+d2+4]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2+4] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+40]  ; cDataDP[j2+k+d2+5] * ExpReal
        fld     qword ptr [eax+ebx+32]  ; cDataDP[j2+k+d2+4]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2+4] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax+32]      ; cDataDP[j2+k+4]
        fsub    st,st(1)                ; cDataDP[j2+k+4] - tempReal
        fstp    qword ptr [eax+ebx+32]  ; cDataDP[j2+k+d2+4]
        fld     qword ptr [eax+40]      ; cDataDP[j2+k+5]
        fsub    st,st(2)                ; cDataDP[j2+k+5] - tempImag
        fstp    qword ptr [eax+ebx+40]  ; cDataDP[j2+k+d2+5]
        fadd    qword ptr [eax+32]      ; cDataDP[j2+k+4] + tempReal
        fstp    qword ptr [eax+32]      ; cDataDP[j2+k+4]
        fadd    qword ptr [eax+40]      ; cDataDP[j2+k+5] + tempImag
        fstp    qword ptr [eax+40]      ; cDataDP[j2+k+5]

        fld     ExpImag4DP              ; PART 4
        fld     qword ptr [eax+ebx+56]  ; cDataDP[j2+k+d2+7]
        fmul    st,st(1)                ; cDataDP[j2+k+d2+7] * ExpImag
        fld     ExpReal4DP
        fld     qword ptr [eax+ebx+48]  ; cDataDP[j2+k+d2+6]        
        fmul    st,st(1)                ; cDataDP[j2+k+d2+6] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    qword ptr [eax+ebx+56]  ; cDataDP[j2+k+d2+7] * ExpReal
        fld     qword ptr [eax+ebx+48]  ; cDataDP[j2+k+d2+6]        
        fmulp   st(3),st                ; cDataDP[j2+k+d2+6] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     qword ptr [eax+48]      ; cDataDP[j2+k+6]
        fsub    st,st(1)                ; cDataDP[j2+k+6] - tempReal
        fstp    qword ptr [eax+ebx+48]  ; cDataDP[j2+k+d2+6]
        fld     qword ptr [eax+56]      ; cDataDP[j2+k+7]
        fsub    st,st(2)                ; cDataDP[j2+k+7] - tempImag
        fstp    qword ptr [eax+ebx+56]  ; cDataDP[j2+k+d2+7]
        fadd    qword ptr [eax+48]      ; cDataDP[j2+k+6] + tempReal
        fstp    qword ptr [eax+48]      ; cDataDP[j2+k+6]
        fadd    qword ptr [eax+56]      ; cDataDP[j2+k+7] + tempImag
        fstp    qword ptr [eax+56]      ; cDataDP[j2+k+7]

        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} 

void ass1SP(int n, int j, int d2,  int cDataSP)
{
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;
    __asm
    {
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP            ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
    lp: mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]

        fld     ExpImag1SP
        fld     dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+1] * ExpImag
        fld     ExpReal1SP
        fld     dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1] * ExpReal
        fld     dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax]         ; cDataSP[j2+k]
        fsub    st,st(1)                ; cDataSP[j2+k] - tempReal
        fstp    dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]
        fld     dword ptr [eax+4]       ; cDataSP[j2+k+1]
        fsub    st,st(2)                ; cDataSP[j2+k+1] - tempImag
        fstp    dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1]
        fadd    dword ptr [eax]         ; cDataSP[j2+k] + tempReal
        fstp    dword ptr [eax]         ; cDataSP[j2+k]
        fadd    dword ptr [eax+4]       ; cDataSP[j2+k+1] + tempImag
        fstp    dword ptr [eax+4]       ; cDataSP[j2+k+1]

        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} 
void ass2SP(int n, int j, int d2,  int cDataSP)
{
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;
    __asm
    {
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP            ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
    lp: mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]

        fld     ExpImag1SP                ; PART 1
        fld     dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+1] * ExpImag
        fld     ExpReal1SP
        fld     dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1] * ExpReal
        fld     dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax]         ; cDataSP[j2+k]
        fsub    st,st(1)                ; cDataSP[j2+k] - tempReal
        fstp    dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]
        fld     dword ptr [eax+4]       ; cDataSP[j2+k+1]
        fsub    st,st(2)                ; cDataSP[j2+k+1] - tempImag
        fstp    dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1]
        fadd    dword ptr [eax]         ; cDataSP[j2+k] + tempReal
        fstp    dword ptr [eax]         ; cDataSP[j2+k]
        fadd    dword ptr [eax+4]       ; cDataSP[j2+k+1] + tempImag
        fstp    dword ptr [eax+4]       ; cDataSP[j2+k+1]

        fld     ExpImag2SP                ; PART 2
        fld     dword ptr [eax+ebx+12]  ; cDataSP[j2+k+d2+3]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+3] * ExpImag
        fld     ExpReal2SP
        fld     dword ptr [eax+ebx+8]   ; cDataSP[j2+k+d2+2]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2+2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+12]  ; cDataSP[j2+k+d2+3] * ExpReal
        fld     dword ptr [eax+ebx+8]   ; cDataSP[j2+k+d2+2]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2+2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax+8]       ; cDataSP[j2+k+2]
        fsub    st,st(1)                ; cDataSP[j2+k+2] - tempReal
        fstp    dword ptr [eax+ebx+8]   ; cDataSP[j2+k+d2+2]
        fld     dword ptr [eax+12]      ; cDataSP[j2+k+3]
        fsub    st,st(2)                ; cDataSP[j2+k+3] - tempImag
        fstp    dword ptr [eax+ebx+12]  ; cDataSP[j2+k+d2+3]
        fadd    dword ptr [eax+8]       ; cDataSP[j2+k+2] + tempReal
        fstp    dword ptr [eax+8]       ; cDataSP[j2+k+2]
        fadd    dword ptr [eax+12]      ; cDataSP[j2+k+3] + tempImag
        fstp    dword ptr [eax+12]      ; cDataSP[j2+k+3]

        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} 

void ass4SP(int n, int j, int d2,  int cDataSP)
{
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;
    __asm
    {
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP              ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]

    lp: fld     ExpImag1SP              ; PART 1
        fld     dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+1] * ExpImag
        fld     ExpReal1SP
        fld     dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1] * ExpReal
        fld     dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax]         ; cDataSP[j2+k]
        fsub    st,st(1)                ; cDataSP[j2+k] - tempReal
        fstp    dword ptr [eax+ebx]     ; cDataSP[j2+k+d2]
        fld     dword ptr [eax+4]       ; cDataSP[j2+k+1]
        fsub    st,st(2)                ; cDataSP[j2+k+1] - tempImag
        fstp    dword ptr [eax+ebx+4]   ; cDataSP[j2+k+d2+1]
        fadd    dword ptr [eax]         ; cDataSP[j2+k] + tempReal
        fstp    dword ptr [eax]         ; cDataSP[j2+k]
        fadd    dword ptr [eax+4]       ; cDataSP[j2+k+1] + tempImag
        fstp    dword ptr [eax+4]       ; cDataSP[j2+k+1]

        fld     ExpImag2SP              ; PART 2
        fld     dword ptr [eax+ebx+12]  ; cDataSP[j2+k+d2+3]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+3] * ExpImag
        fld     ExpReal2SP
        fld     dword ptr [eax+ebx+8]   ; cDataSP[j2+k+d2+2]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2+2] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+12]  ; cDataSP[j2+k+d2+3] * ExpReal
        fld     dword ptr [eax+ebx+8]   ; cDataSP[j2+k+d2+2]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2+2] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax+8]       ; cDataSP[j2+k+2]
        fsub    st,st(1)                ; cDataSP[j2+k+2] - tempReal
        fstp    dword ptr [eax+ebx+8]   ; cDataSP[j2+k+d2+2]
        fld     dword ptr [eax+12]      ; cDataSP[j2+k+3]
        fsub    st,st(2)                ; cDataSP[j2+k+3] - tempImag
        fstp    dword ptr [eax+ebx+12]  ; cDataSP[j2+k+d2+3]
        fadd    dword ptr [eax+8]       ; cDataSP[j2+k+2] + tempReal
        fstp    dword ptr [eax+8]       ; cDataSP[j2+k+2]
        fadd    dword ptr [eax+12]      ; cDataSP[j2+k+3] + tempImag
        fstp    dword ptr [eax+12]      ; cDataSP[j2+k+3]

        fld     ExpImag3SP              ; PART 3
        fld     dword ptr [eax+ebx+20]  ; cDataSP[j2+k+d2+5]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+5] * ExpImag
        fld     ExpReal3SP
        fld     dword ptr [eax+ebx+16]  ; cDataSP[j2+k+d2+4]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2+4] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+20]  ; cDataSP[j2+k+d2+5] * ExpReal
        fld     dword ptr [eax+ebx+16]  ; cDataSP[j2+k+d2+4]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2+4] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax+16]      ; cDataSP[j2+k+4]
        fsub    st,st(1)                ; cDataSP[j2+k+4] - tempReal
        fstp    dword ptr [eax+ebx+16]  ; cDataSP[j2+k+d2+4]
        fld     dword ptr [eax+20]      ; cDataSP[j2+k+5]
        fsub    st,st(2)                ; cDataSP[j2+k+5] - tempImag
        fstp    dword ptr [eax+ebx+20]  ; cDataSP[j2+k+d2+5]
        fadd    dword ptr [eax+16]      ; cDataSP[j2+k+4] + tempReal
        fstp    dword ptr [eax+16]      ; cDataSP[j2+k+4]
        fadd    dword ptr [eax+20]      ; cDataSP[j2+k+5] + tempImag
        fstp    dword ptr [eax+20]      ; cDataSP[j2+k+5]

        fld     ExpImag4SP              ; PART 4
        fld     dword ptr [eax+ebx+28]  ; cDataSP[j2+k+d2+7]
        fmul    st,st(1)                ; cDataSP[j2+k+d2+7] * ExpImag
        fld     ExpReal4SP
        fld     dword ptr [eax+ebx+24]  ; cDataSP[j2+k+d2+6]        
        fmul    st,st(1)                ; cDataSP[j2+k+d2+6] * ExpReal
        fsubrp  st(2),st                ; tempReal
        fmul    dword ptr [eax+ebx+28]  ; cDataSP[j2+k+d2+7] * ExpReal
        fld     dword ptr [eax+ebx+24]  ; cDataSP[j2+k+d2+6]        
        fmulp   st(3),st                ; cDataSP[j2+k+d2+6] *  ExpImag
        faddp   st(2),st                ; tempImag
        fld     dword ptr [eax+24]      ; cDataSP[j2+k+6]
        fsub    st,st(1)                ; cDataSP[j2+k+6] - tempReal
        fstp    dword ptr [eax+ebx+24]  ; cDataSP[j2+k+d2+6]
        fld     dword ptr [eax+28]      ; cDataSP[j2+k+7]
        fsub    st,st(2)                ; cDataSP[j2+k+7] - tempImag
        fstp    dword ptr [eax+ebx+28]  ; cDataSP[j2+k+d2+7]
        fadd    dword ptr [eax+24]      ; cDataSP[j2+k+6] + tempReal
        fstp    dword ptr [eax+24]      ; cDataSP[j2+k+6]
        fadd    dword ptr [eax+28]      ; cDataSP[j2+k+7] + tempImag
        fstp    dword ptr [eax+28]      ; cDataSP[j2+k+7]

        add     eax, edx                ; address cDataDP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} 

void ass2SPSSE(int n, int j, int d2,  int cDataSP)
{
     
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;

    cExpSPA[0] =   ExpImag1SP;
    cExpSPA[1] = - ExpImag1SP;
    cExpSPA[2] =   ExpImag2SP;
    cExpSPA[3] = - ExpImag2SP;
    cExpSPA[4] =   ExpReal1SP;
    cExpSPA[5] =   ExpReal1SP;
    cExpSPA[6] =   ExpReal2SP;
    cExpSPA[7] =   ExpReal2SP;

    __asm
    {
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP            ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]
        mov     esi, cExpSPA
        movaps  xmm2, [esi]             ; ExpImag1SP, - ExpImag1SP
        movaps  xmm3, [esi+16]          ; ExpReal1SP,   ExpReal1SP

        align 16
    lp: movaps  xmm0, [eax+ebx]         ; cDataSP[j2+k+d2] & [j2+k+d2+1] & [j2+k+d2+2] & [j2+k+d2+3]
        movaps  xmm1, xmm0              ; cDataSP[j2+k+d2] & [j2+k+d2+1] & [j2+k+d2+2] & [j2+k+d2+3]
        mulps   xmm0, xmm2              ; cDataSP[pair] * ExpImag1SP, -ExpImag1SP
        shufps  xmm0, xmm0, 177         ; reverse 2 pairs
        mulps   xmm1, xmm3              ; cDataSP[quad] * ExpReal1SP, ExpReal1SP
        addps   xmm1, xmm0              ; tempReal1, tempImag1, tempReal2, tempImag2
        movaps  xmm4, [eax]             ; cDataSP[j2+k] & [j2+k+1] & [j2+k+2] & [j2+k+3]
        movaps  xmm5, xmm4              ; cDataSP[j2+k] & [j2+k+1] & [j2+k+2] & [j2+k+3]
        subps   xmm4, xmm1              ; cDataSP[pair] - tempReal, tempImag 
        movaps  [eax+ebx], xmm4         ; save cDataSP[j2+k+d2] & [j2+k+d2+1] & [j2+k+d2+2] & [j2+k+d2+3]
        addps   xmm5, xmm1              ; cDataSP[quad] + tempReal, tempImag
        movaps  [eax], xmm5             ; save cDataSP[j2+k] & cDataSP[j2+k+1]  & [j2+k+2] & [j2+k+3]
        add     eax, edx                ; address cDataSP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} // ass2SPSSE

void ass4SPSSE(int n, int j, int d2,  int cDataSP)
{     
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;

    cExpSPA[0]  =   ExpImag1SP;
    cExpSPA[1]  = - ExpImag1SP;
    cExpSPA[2]  =   ExpImag2SP;
    cExpSPA[3]  = - ExpImag2SP;
    cExpSPA[4]  =   ExpImag3SP;
    cExpSPA[5]  = - ExpImag3SP;
    cExpSPA[6]  =   ExpImag4SP;
    cExpSPA[7]  = - ExpImag4SP;
    cExpSPA[8]  =   ExpReal1SP;
    cExpSPA[9]  =   ExpReal1SP;
    cExpSPA[10] =   ExpReal2SP;
    cExpSPA[11] =   ExpReal2SP;
    cExpSPA[12] =   ExpReal3SP;
    cExpSPA[13] =   ExpReal3SP;
    cExpSPA[14] =   ExpReal4SP;
    cExpSPA[15] =   ExpReal4SP;
    
    __asm
    {
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP            ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]
        mov     esi, cExpSPA
        movaps  xmm4, [esi]             ; ExpImag1SP, - ExpImag1SP, ExpImag2SP, - ExpImag2SP
        movaps  xmm5, [esi+16]          ; ExpImag3SP, - ExpImag3SP, ExpImag4SP, - ExpImag4SP
        movaps  xmm6, [esi+32]          ; ExpReal1SP,   ExpReal1SP, ExpReal2SP,   ExpReal2SP
        movaps  xmm7, [esi+48]          ; ExpReal3SP,   ExpReal3SP, ExpReal4SP,   ExpReal4SP
        align 16
    lp: movaps  xmm0, [eax+ebx]         ; cDataSP[j2+k+d2]   & [j2+k+d2+1] & [j2+k+d2+2] & [j2+k+d2+3]
        movaps  xmm1, [eax+ebx+16]      ; cDataSP[j2+k+d2+4] & [j2+k+d2+5] & [j2+k+d2+6] & [j2+k+d2+7]
        movaps  xmm2, xmm0              ; cDataSP[j2+k+d2]   & [j2+k+d2+1] & [j2+k+d2+2] & [j2+k+d2+3]
        movaps  xmm3, xmm1              ; cDataSP[j2+k+d2+2] & [j2+k+d2+3] & [j2+k+d2+6] & [j2+k+d2+7]
        mulps   xmm0, xmm4              ; cDataSP[1] * ExpImag1SP, - ExpImag1SP, ExpImag2SP, - ExpImag2SP
        mulps   xmm1, xmm5              ; cDataSP[2] * ExpImag3SP, - ExpImag3SP, ExpImag4SP, - ExpImag4SP
        shufps  xmm0, xmm0, 177         ; reverse 2 pairs
        shufps  xmm1, xmm1, 177         ; reverse 2 pairs
        mulps   xmm2, xmm6              ; cDataSP[1] * ExpReal1SP, ExpReal1SP, ExpReal2SP, ExpReal2SP
        mulps   xmm3, xmm7              ; cDataSP[2] * ExpReal3SP, ExpReal3SP, ExpReal4SP, ExpReal4SP
        addps   xmm0, xmm2              ; tempReal1, tempImag1, tempReal2, tempImag2
        addps   xmm1, xmm3              ; tempReal3, tempImag3, tempReal4, tempImag4
        movaps  xmm2, [eax]             ; cDataSP[j2+k]   & [j2+k+1] & [j2+k+2] & [j2+k+3]
        movaps  xmm3, [eax+16]          ; cDataSP[j2+k+4] & [j2+k+5] & [j2+k+6] & [j2+k+7]
        subps   xmm2, xmm0              ; cDataSP[quad] - tempReal1, tempImag1, tempReal2, tempImag2
        subps   xmm3, xmm1              ; cDataSP[quad] - tempReal3, tempImag3, tempReal4, tempImag4
        movaps  [eax+ebx],    xmm2      ; save cDataSP[j2+k+d2]   & [j2+k+d2+1] & [j2+k+d2+2] & [j2+k+d2+3]
        movaps  [eax+ebx+16], xmm3      ; save cDataSP[j2+k+d2+4] & [j2+k+d2+5] & [j2+k+d2+6] & [j2+k+d2+7]
        movaps  xmm2, [eax]             ; cDataSP[j2+k]   & [j2+k+1] & [j2+k+2] & [j2+k+3]
        movaps  xmm3, [eax+16]          ; cDataSP[j2+k+4] & [j2+k+5] & [j2+k+6] & [j2+k+7]
        addps   xmm2, xmm0              ; cDataSP[quad] + tempReal1, tempImag1, tempReal2, tempImag2
        addps   xmm3, xmm1              ; cDataSP[quad] + tempReal3, tempImag3, tempReal4, tempImag4
        movaps  [eax],    xmm2          ; save cDataSP[j2+k]   & [j2+k+1] & [j2+k+2] & [j2+k+3]
        movaps  [eax+16], xmm3          ; save cDataSP[j2+k+4] & [j2+k+5] & [j2+k+6] & [j2+k+7]
        add     eax, edx                ; address cDataSP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} // ass4SPSSE


void ass1DPSSE2(int n, int j, int d2,  int cDataDP)
{
     
    n = n * 8;
    j = j * 8;
    d2 = d2 * 8;

    cExpDPA[0] =   ExpImag1DP;
    cExpDPA[1] = - ExpImag1DP;
    cExpDPA[2] =   ExpReal1DP;
    cExpDPA[3] =   ExpReal1DP;

    __asm
    {
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataDP            ; not lea
        add     esi, edx                ; address cDataDP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataDP[j2]
        add     eax, ecx                ; address cDataDP[jk or j2 + k]
        mov     esi, cExpDPA
        movapd  xmm2, [esi]             ; ExpImag1DP, - ExpImag1DP
        movapd  xmm3, [esi+16]          ; ExpReal1DP,   ExpReal1DP

        align 16
    lp: movapd  xmm4, [eax]             ; cDataDP[j2+k] & cDataDP[j2+k+1]
        movapd  xmm0, [eax+ebx]         ; cDataDP[j2+k+d2] & [j2+k+d2+1]
        movapd  xmm5, xmm4              ; cDataDP[j2+k] & cDataDP[j2+k+1]
        movapd  xmm1, xmm0              ; cDataDP[j2+k+d2] & [j2+k+d2+1]
        mulpd   xmm0, xmm2              ; cDataDP[pair] * ExpImag1DP, -ExpImag1DP
        shufpd  xmm0, xmm0, 1           ; reverse pair
        mulpd   xmm1, xmm3              ; cDataDP[pair] * ExpReal1DP, ExpReal1DP
        addpd   xmm1, xmm0              ; tempReal, tempImag
        subpd   xmm4, xmm1              ; cDataDP[pair] - tempReal, tempImag 
        addpd   xmm5, xmm1              ; cDataDP[pair] + tempReal, tempImag
        movapd  [eax], xmm5             ; save cDataDP[j2+k] & cDataDP[j2+k+1]
        movapd  [eax+ebx], xmm4         ; save cDataDP[j2+k+d2] & [j2+k+d2+1]

        add     eax, edx                ; address cDataDP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} // ass1DPSSE2


void ass2DPSSE2(int n, int j, int d2,  int cDataDP)
{     
    n = n * 8;
    j = j * 8;
    d2 = d2 * 8;

    cExpDPA[0] =   ExpImag1DP;
    cExpDPA[1] = - ExpImag1DP;
    cExpDPA[2] =   ExpImag2DP;
    cExpDPA[3] = - ExpImag2DP;
    cExpDPA[4] =   ExpReal1DP;
    cExpDPA[5] =   ExpReal1DP;
    cExpDPA[6] =   ExpReal2DP;
    cExpDPA[7] =   ExpReal2DP;
    
    __asm
    {
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataDP            ; not lea
        add     esi, edx                ; address cDataDP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataDP[j2]
        add     eax, ecx                ; address cDataDP[jk or j2 + k]
        mov     esi, cExpDPA
        movapd  xmm4, [esi]             ; ExpImag1DP, - ExpImag1DP
        movapd  xmm5, [esi+16]          ; ExpImag2DP, - ExpImag2DP
        movapd  xmm6, [esi+32]          ; ExpReal1DP,   ExpReal1DP
        movapd  xmm7, [esi+48]          ; ExpReal2DP,   ExpReal2DP
        align 16
    lp: movapd  xmm0, [eax+ebx]         ; cDataDP[j2+k+d2]   & [j2+k+d2+1]
        movapd  xmm1, [eax+ebx+16]      ; cDataDP[j2+k+d2+2] & [j2+k+d2+3]
        movapd  xmm2, xmm0              ; cDataDP[j2+k+d2]   & [j2+k+d2+1]
        movapd  xmm3, xmm1              ; cDataDP[j2+k+d2+2] & [j2+k+d2+3]
        mulpd   xmm0, xmm4              ; ExpImag1DP, - ExpImag1DP
        mulpd   xmm1, xmm5              ; ExpImag2DP, - ExpImag2DP
        shufpd  xmm0, xmm0, 1           ; reverse pair
        shufpd  xmm1, xmm1, 1           ; reverse pair
        mulpd   xmm2, xmm6              ; ExpReal1DP, ExpReal1DP
        mulpd   xmm3, xmm7              ; ExpReal2DP, ExpReal2DP
        addpd   xmm0, xmm2              ; tempReal1, tempImag1
        addpd   xmm1, xmm3              ; tempReal2, tempImag2
        movapd  xmm2, [eax]             ; cDataDP[j2+k]   & cDataDP[j2+k+1]
        movapd  xmm3, [eax+16]          ; cDataDP[j2+k+2] & cDataDP[j2+k+3]
        subpd   xmm2, xmm0              ; cDataDP[pair] - tempReal1, tempImag1
        subpd   xmm3, xmm1              ; cDataDP[pair] - tempReal2, tempImag2
        movapd  [eax+ebx],    xmm2      ; save cDataDP[j2+k+d2]   & [j2+k+d2+1]
        movapd  [eax+ebx+16], xmm3      ; save cDataDP[j2+k+d2+2] & [j2+k+d2+3]
        movapd  xmm2, [eax]             ; cDataDP[j2+k]   & cDataDP[j2+k+1]
        movapd  xmm3, [eax+16]          ; cDataDP[j2+k+2] & cDataDP[j2+k+3]
        addpd   xmm2, xmm0              ; cDataDP[pair] + tempReal1, tempImag1
        addpd   xmm3, xmm1              ; cDataDP[pair] + tempReal2, tempImag2
        movapd  [eax],    xmm2          ; save cDataDP[j2+k]   & [j2+k+1]
        movapd  [eax+16], xmm3          ; save cDataDP[j2+k+2] & [j2+kd+3]
        add     eax, edx                ; address cDataDP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} // ass2DPSSE2

void ass4DPSSE2(int n, int j, int d2,  int cDataDP)
{
     
    n = n * 8;
    j = j * 8;
    d2 = d2 * 8;

    cExpDPA[0]  =   ExpImag1DP;
    cExpDPA[1]  = - ExpImag1DP;
    cExpDPA[2]  =   ExpImag2DP;
    cExpDPA[3]  = - ExpImag2DP;
    cExpDPA[4]  =   ExpImag3DP;
    cExpDPA[5]  = - ExpImag3DP;
    cExpDPA[6]  =   ExpImag4DP;
    cExpDPA[7]  = - ExpImag4DP;
    cExpDPA[8]  =   ExpReal1DP;
    cExpDPA[9]  =   ExpReal1DP;
    cExpDPA[10] =   ExpReal2DP;
    cExpDPA[11] =   ExpReal2DP;
    cExpDPA[12] =   ExpReal3DP;
    cExpDPA[13] =   ExpReal3DP;
    cExpDPA[14] =   ExpReal4DP;
    cExpDPA[15] =   ExpReal4DP;
    
    __asm
    {
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataDP            ; not lea
        add     esi, edx                ; address cDataDP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataDP[j2]
        add     eax, ecx                ; address cDataDP[jk or j2 + k]
        mov     esi, cExpDPA
        align 16
    lp:
        movapd  xmm0, [eax+ebx]         ; cDataDP[j2+k+d2]   & [j2+k+d2+1]
        movapd  xmm1, [eax+ebx+16]      ; cDataDP[j2+k+d2+2] & [j2+k+d2+3]
        movapd  xmm2, [eax+ebx+32]      ; cDataDP[j2+k+d2+4] & [j2+k+d2+5]
        movapd  xmm3, [eax+ebx+48]      ; cDataDP[j2+k+d2+6] & [j2+k+d2+7]
        mulpd   xmm0, [esi]             ; ExpImag1DP, - ExpImag1DP
        mulpd   xmm1, [esi+16]          ; ExpImag2DP, - ExpImag2DP
        mulpd   xmm2, [esi+32]          ; ExpImag3DP, - ExpImag3DP
        mulpd   xmm3, [esi+48]          ; ExpImag4DP, - ExpImag4DP
        shufpd  xmm0, xmm0, 1           ; reverse pair
        shufpd  xmm1, xmm1, 1           ; reverse pair
        shufpd  xmm2, xmm2, 1           ; reverse pair
        shufpd  xmm3, xmm3, 1           ; reverse pair
        movapd  xmm4, [eax+ebx]         ; cDataDP[j2+k+d2]   & [j2+k+d2+1]
        movapd  xmm5, [eax+ebx+16]      ; cDataDP[j2+k+d2+2] & [j2+k+d2+3]
        movapd  xmm6, [eax+ebx+32]      ; cDataDP[j2+k+d2+4] & [j2+k+d2+5]
        movapd  xmm7, [eax+ebx+48]      ; cDataDP[j2+k+d2+6] & [j2+k+d2+7]
        mulpd   xmm4, [esi+64]          ; ExpReal1DP, ExpReal1DP
        mulpd   xmm5, [esi+80]          ; ExpReal2DP, ExpReal2DP
        mulpd   xmm6, [esi+96]          ; ExpReal3DP, ExpReal3DP
        mulpd   xmm7, [esi+112]         ; ExpReal4DP, ExpReal4DP
        addpd   xmm0, xmm4              ; tempReal1, tempImag1
        addpd   xmm1, xmm5              ; tempReal2, tempImag2
        addpd   xmm2, xmm6              ; tempReal3, tempImag3
        addpd   xmm3, xmm7              ; tempReal4, tempImag4
        movapd  xmm4, [eax]             ; cDataDP[j2+k]   & cDataDP[j2+k+1]
        movapd  xmm5, [eax+16]          ; cDataDP[j2+k+2] & cDataDP[j2+k+3]
        movapd  xmm6, [eax+32]          ; cDataDP[j2+k+4] & cDataDP[j2+k+5]
        movapd  xmm7, [eax+48]          ; cDataDP[j2+k+6] & cDataDP[j2+k+7]
        subpd   xmm4, xmm0              ; cDataDP[pair] - tempReal1, tempImag1
        subpd   xmm5, xmm1              ; cDataDP[pair] - tempReal2, tempImag2
        subpd   xmm6, xmm2              ; cDataDP[pair] - tempReal3, tempImag3
        subpd   xmm7, xmm3              ; cDataDP[pair] - tempReal4, tempImag4
        movapd  [eax+ebx],    xmm4      ; save cDataDP[j2+k+d2]   & [j2+k+d2+1]
        movapd  [eax+ebx+16], xmm5      ; save cDataDP[j2+k+d2+2] & [j2+k+d2+3]
        movapd  [eax+ebx+32], xmm6      ; save cDataDP[j2+k+d2+4] & [j2+k+d2+5]
        movapd  [eax+ebx+48], xmm7      ; save cDataDP[j2+k+d2+6] & [j2+k+d2+7]
        movapd  xmm4, [eax]             ; cDataDP[j2+k]   & cDataDP[j2+k+1]
        movapd  xmm5, [eax+16]          ; cDataDP[j2+k+2] & cDataDP[j2+k+3]
        movapd  xmm6, [eax+32]          ; cDataDP[j2+k+4] & cDataDP[j2+k+5]
        movapd  xmm7, [eax+48]          ; cDataDP[j2+k+6] & cDataDP[j2+k+7]
        addpd   xmm4, xmm0              ; cDataDP[pair] + tempReal1, tempImag1
        addpd   xmm5, xmm1              ; cDataDP[pair] + tempReal2, tempImag2
        addpd   xmm6, xmm2              ; cDataDP[pair] + tempReal3, tempImag3
        addpd   xmm7, xmm3              ; cDataDP[pair] + tempReal4, tempImag4
        movapd  [eax],    xmm4          ; save cDataDP[j2+k]   & [j2+k+1]
        movapd  [eax+16], xmm5          ; save cDataDP[j2+k+2] & [j2+kd+3]
        movapd  [eax+32], xmm6          ; save cDataDP[j2+k+4] & [j2+k+5]
        movapd  [eax+48], xmm7          ; save cDataDP[j2+k+6] & [j2+k+7]
        add     eax, edx                ; address cDataDP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
     }
} // ass4DPSSE2

            
void ass1SP3DNow(int n, int j, int d2,  int cDataSP)
{
     
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;

    cExpSPA[0] =   ExpImag1SP;
    cExpSPA[1] = - ExpImag1SP;
    cExpSPA[2] =   ExpReal1SP;
    cExpSPA[3] =   ExpReal1SP;

    __asm
    {
        femms 
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP            ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]
        mov     esi, cExpSPA
        movq    mm2, [esi]              ; ExpImag1SP, - ExpImag1SP
        movq    mm3, [esi+8]            ; ExpReal1SP,   ExpReal1SP

        align 16
    lp: movq    mm4, [eax]              ; cDataSP[j2+k] & cDataSP[j2+k+1]
        movq    mm0, [eax+8]            ; cDataSP[j2+k+d2] & [j2+k+d2+1]
        movq    mm1, mm0                ; cDataSP[j2+k+d2] & [j2+k+d2+1]
        pfmul   mm0, mm2                ; cDataSP[pair] * ExpImag1SP, -ExpImag1SP
        pswapd  mm0, mm0                ; reverse pair
        pfmul   mm1, mm3                ; cDataSP[pair] * ExpReal1SP, ExpReal1SP
        pfadd   mm1, mm0                ; tempReal, tempImag
        movq    mm5, mm4                ; cDataSP[j2+k] & cDataSP[j2+k+1]
        pfsub   mm4, mm1                ; cDataSP[pair] - tempReal, tempImag 
        pfadd   mm5, mm1                ; cDataSP[pair] + tempReal, tempImag
        movq    [eax], mm5              ; save cDataSP[j2+k] & cDataSP[j2+k+1]
        movq    [eax+8], mm4            ; save cDataSP[j2+k+d2] & [j2+k+d2+1]

        add     eax, edx                ; address cDataSP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
        femms 
     }
} // ass1SP3DNow


void ass2SP3DNow(int n, int j, int d2,  int cDataSP)
{
     
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;

    cExpSPA[0] =   ExpImag1SP;
    cExpSPA[1] = - ExpImag1SP;
    cExpSPA[2] =   ExpReal1SP;
    cExpSPA[3] =   ExpReal1SP;
    cExpSPA[4] =   ExpImag2SP;
    cExpSPA[5] = - ExpImag2SP;
    cExpSPA[6] =   ExpReal2SP;
    cExpSPA[7] =   ExpReal2SP;
    
    __asm
    {
        femms 
        align 16
        mov     ebx, d2
        mov     ecx, 0                  ; k
        mov     edx, j
        add     edx, edx                ; j*2
        mov     edi, n
        add     edi, edi                ; n*2
        mov     esi, cDataSP            ; not lea
        add     esi, edx                ; address cDataSP[j2]
        mov     edx, ebx                ; d2
        add     edx, edx                ; d2*2
        mov     eax, esi                ; address cDataSP[j2]
        add     eax, ecx                ; address cDataSP[jk or j2 + k]
        mov     esi, cExpSPA
        movq    mm2, [esi]              ; ExpImag1SP, - ExpImag1SP
        movq    mm3, [esi+8]            ; ExpReal1SP,   ExpReal1SP
        movq    mm6, [esi+16]           ; ExpImag2SP, - ExpImag2SP
        movq    mm7, [esi+24]           ; ExpReal2SP,   ExpReal2SP

        align 16
    lp: movq    mm0, [eax+ebx]          ; cDataSP[j2+k+d2] & [j2+k+d2+1]
        movq    mm1, mm0                ; cDataSP[j2+k+d2] & [j2+k+d2+1]
        pfmul   mm0, mm2                ; cDataSP[pair] * ExpImag1SP, -ExpImag1SP
        pswapd  mm0, mm0                ; reverse pair
        pfmul   mm1, mm3                ; cDataSP[pair] * ExpReal1SP, ExpReal1SP
        pfadd   mm1, mm0                ; tempReal, tempImag
        movq    mm4, [eax]              ; cDataSP[j2+k] & cDataSP[j2+k+1]
        movq    mm5, mm4                ; cDataSP[j2+k] & cDataSP[j2+k+1]
        pfsub   mm4, mm1                ; cDataSP[pair] - tempReal, tempImag 
        movq    [eax+ebx], mm4          ; save cDataSP[j2+k+d2] & [j2+k+d2+1]
        pfadd   mm5, mm1                ; cDataSP[pair] + tempReal, tempImag
        movq    [eax], mm5              ; save cDataSP[j2+k] & cDataSP[j2+k+1]

        movq    mm0, [eax+ebx+8]        ; cDataSP[j2+k+d2+2] & [j2+k+d2+3]
        movq    mm1, mm0                ; cDataSP[j2+k+d2+2] & [j2+k+d2+3]
        pfmul   mm0, mm6                ; cDataSP[pair] * ExpImag2SP, -ExpImag2SP
        pswapd  mm0, mm0                ; reverse pair
        pfmul   mm1, mm7                ; cDataSP[pair] * ExpReal2SP, ExpReal2SP
        pfadd   mm1, mm0                ; tempReal, tempImag
        movq    mm4, [eax+8]            ; cDataSP[j2+k+2] & cDataSP[j2+k+3]
        movq    mm5, mm4                ; cDataSP[j2+k+2] & cDataSP[j2+k+3]
        pfsub   mm4, mm1                ; cDataSP[pair] - tempReal, tempImag 
        movq    [eax+ebx+8], mm4        ; save cDataSP[j2+k+d2+2] & [j2+k+d2+3]
        pfadd   mm5, mm1                ; cDataSP[pair] + tempReal, tempImag
        movq    [eax+8], mm5            ; save cDataSP[j2+k+2] & cDataSP[j2+k+3]

        add     eax, edx                ; address cDataSP[j2+k+d2]
        add     ecx, edx                ; k=k+d2
        cmp     ecx, edi                ; k vs b2
        jl      lp
        femms 
     }
} // ass2SP3DNow

void ass4SP3DNow(int n, int j, int d2,  int cDataSP)
{
     
    n = n * 4;
    j = j * 4;
    d2 = d2 * 4;

    cExpSPA[0]  =   ExpImag1SP;
    cExpSPA[1]  = - ExpImag1SP;
    cExpSPA[2]  =   ExpImag2SP;
    cExpSPA[3]  = - ExpImag2SP;
    cExpSPA[4]  =   ExpImag3SP;
    cExpSPA[5]  = - ExpImag3SP;
    cExpSPA[6]  =   ExpImag4SP;
    cExpSPA[7]  = - ExpImag4SP;
    cExpSPA[8]  =   ExpReal1SP;
    cExpSPA[9]  =   ExpReal1SP;
    cExpSPA[10] =   ExpReal2SP;
    cExpSPA[11] =   ExpReal2SP;
    cExpSPA[12] =   ExpReal3SP;
    cExpSPA[13] =   ExpReal3SP;
    cExpSPA[14] =   ExpReal4SP;
    cExpSPA[15] =   ExpReal4SP;
    
    __asm
    {
        femms 
        align 16
        mov     ebx, d2
        mov     ecx, 0               ; k
        mov     edx, j
        add     edx, edx             ; j*2
        mov     edi, n
        add     edi, edi             ; n*2
        mov     esi, cDataSP         ; not lea
        add     esi, edx             ; address cDataSP[j2]
        mov     edx, ebx             ; d2
        add     edx, edx             ; d2*2
        mov     eax, esi             ; address cDataSP[j2]
        add     eax, ecx             ; address cDataSP[jk or j2 + k]
        mov     esi, cExpSPA
        align 16
    lp: movq  mm0, [eax+ebx]         ; cDataSP[j2+k+d2]   & [j2+k+d2+1]
        movq  mm1, [eax+ebx+8]       ; cDataSP[j2+k+d2+2] & [j2+k+d2+3]
        movq  mm2, [eax+ebx+16]      ; cDataSP[j2+k+d2+4] & [j2+k+d2+5]
        movq  mm3, [eax+ebx+24]      ; cDataSP[j2+k+d2+6] & [j2+k+d2+7]
        pfmul   mm0, [esi]           ; ExpImag1SP, - ExpImag1SP
        pfmul   mm1, [esi+8]         ; ExpImag2SP, - ExpImag2SP
        pfmul   mm2, [esi+16]        ; ExpImag3SP, - ExpImag3SP
        pfmul   mm3, [esi+24]        ; ExpImag4SP, - ExpImag4SP
        pswapd  mm0, mm0             ; reverse pair
        pswapd  mm1, mm1             ; reverse pair
        pswapd  mm2, mm2             ; reverse pair
        pswapd  mm3, mm3             ; reverse pair
        movq  mm4, [eax+ebx]         ; cDataSP[j2+k+d2]   & [j2+k+d2+1]
        movq  mm5, [eax+ebx+8]       ; cDataSP[j2+k+d2+2] & [j2+k+d2+3]
        movq  mm6, [eax+ebx+16]      ; cDataSP[j2+k+d2+4] & [j2+k+d2+5]
        movq  mm7, [eax+ebx+24]      ; cDataSP[j2+k+d2+6] & [j2+k+d2+7]
        pfmul   mm4, [esi+32]        ; ExpReal1SP, ExpReal1SP
        pfmul   mm5, [esi+40]        ; ExpReal2SP, ExpReal2SP
        pfmul   mm6, [esi+48]        ; ExpReal3SP, ExpReal3SP
        pfmul   mm7, [esi+56]        ; ExpReal4SP, ExpReal4SP
        pfadd   mm0, mm4             ; tempReal1, tempImag1
        pfadd   mm1, mm5             ; tempReal2, tempImag2
        pfadd   mm2, mm6             ; tempReal3, tempImag3
        pfadd   mm3, mm7             ; tempReal4, tempImag4
        movq  mm4, [eax]             ; cDataSP[j2+k]   & cDataSP[j2+k+1]
        movq  mm5, [eax+8]           ; cDataSP[j2+k+2] & cDataSP[j2+k+3]
        movq  mm6, [eax+16]          ; cDataSP[j2+k+4] & cDataSP[j2+k+5]
        movq  mm7, [eax+24]          ; cDataSP[j2+k+6] & cDataSP[j2+k+7]
        pfsub   mm4, mm0             ; cDataSP[pair] - tempReal1, tempImag1
        pfsub   mm5, mm1             ; cDataSP[pair] - tempReal2, tempImag2
        pfsub   mm6, mm2             ; cDataSP[pair] - tempReal3, tempImag3
        pfsub   mm7, mm3             ; cDataSP[pair] - tempReal4, tempImag4
        movq  [eax+ebx],    mm4      ; save cDataSP[j2+k+d2]   & [j2+k+d2+1]
        movq  [eax+ebx+8], mm5       ; save cDataSP[j2+k+d2+2] & [j2+k+d2+3]
        movq  [eax+ebx+16], mm6      ; save cDataSP[j2+k+d2+4] & [j2+k+d2+5]
        movq  [eax+ebx+24], mm7      ; save cDataSP[j2+k+d2+6] & [j2+k+d2+7]
        movq  mm4, [eax]             ; cDataSP[j2+k]   & cDataSP[j2+k+1]
        movq  mm5, [eax+8]           ; cDataSP[j2+k+2] & cDataSP[j2+k+3]
        movq  mm6, [eax+16]          ; cDataSP[j2+k+4] & cDataSP[j2+k+5]
        movq  mm7, [eax+24]          ; cDataSP[j2+k+6] & cDataSP[j2+k+7]
        pfadd   mm4, mm0             ; cDataSP[pair] + tempReal1, tempImag1
        pfadd   mm5, mm1             ; cDataSP[pair] + tempReal2, tempImag2
        pfadd   mm6, mm2             ; cDataSP[pair] + tempReal3, tempImag3
        pfadd   mm7, mm3             ; cDataSP[pair] + tempReal4, tempImag4
        movq  [eax],    mm4          ; save cDataSP[j2+k]   & [j2+k+1]
        movq  [eax+8], mm5           ; save cDataSP[j2+k+2] & [j2+kd+3]
        movq  [eax+16], mm6          ; save cDataSP[j2+k+4] & [j2+k+5]
        movq  [eax+24], mm7          ; save cDataSP[j2+k+6] & [j2+k+7]
        add     eax, edx             ; address cDataSP[j2+k+d2]
        add     ecx, edx             ; k=k+d2
        cmp     ecx, edi             ; k vs b2
        jl      lp
        femms 
     }
} // ass4SP3DNow
            
/*

Version 1 was:

    d = 1;
    index = 1;
    while (index <= t)
    {
      d2 = d + d;
      for (j=0; j<d; j++)
      {
        k = 0;
        jnd2 =  n / d2 * j; 
        while (k < n)
        {
          jk = j + k;
          jkd = jk + d;
          tempReal   = (cDataSP[jkd]   * cExpSP[jnd2]) - (cDataSP[jkd+1] * cExpSP[jnd2+1]);
          tempImag   = (cDataSP[jkd+1] * cExpSP[jnd2]) + (cDataSP[jkd]   * cExpSP[jnd2+1]);
          cDataSP[jkd]    = cDataSP[jk]   - tempReal;
          cDataSP[jkd+1]  = cDataSP[jk+1] - tempImag;
          cDataSP[jk]     = cDataSP[jk]   + tempReal;
          cDataSP[jk+1]   = cDataSP[jk+1] + tempImag;
          k = k + d2;
        }
      }
      index = index + 1;
      d = d2;
    }

*/
